# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pytest import mark

from .stack_hasher import (
    GdbStackFrame,
    MinidumpStackFrame,
    RrStackFrame,
    RustStackFrame,
    SanitizerStackFrame,
    Stack,
    StackFrame,
    ThreadSanitizerStackFrame,
    ValgrindStackFrame,
)


class BasicStackFrame(StackFrame):
    @classmethod
    def from_line(cls, input_line):
        # unused
        pass


def test_stack_01():
    """test creating an empty Stack"""
    stack = Stack([])
    assert stack.minor is None
    assert isinstance(stack.frames, list)
    assert stack._major_depth > 0  # pylint: disable=protected-access


def test_stack_02():
    """test creating a Stack with 1 generic frame"""
    frames = [BasicStackFrame(function="a", location="b", offset="c", stack_line="0")]
    stack = Stack(frames)
    assert stack.minor is not None
    assert stack.major is not None
    # at this point the hashes should match
    assert stack.minor == stack.major
    assert len(stack.frames) == 1
    output = str(stack)
    assert "00" in output
    assert "function: 'a'" in output
    assert "location: 'b'" in output
    assert "offset: 'c'" in output


def test_stack_03():
    """test creating a Stack with 2 frames"""
    frames = [
        BasicStackFrame(function="a", location="b", offset="c", stack_line="0")
        for _ in range(2)
    ]
    stack = Stack(frames, major_depth=2)
    assert stack.minor is not None
    assert stack.major is not None
    # at this point the hashes should not match because offset on the major hash is
    # only added from the top frame
    assert stack.minor != stack.major
    assert len(stack.frames) == 2


def test_stack_04():
    """test creating a Stack with 2 frames with a major depth of 0"""
    frames = [
        BasicStackFrame(function="a", location="b", offset="c", stack_line=str(line))
        for line in range(2)
    ]
    stack = Stack(frames, major_depth=0)
    assert stack.minor is not None
    assert stack.major is None
    # at this point the hashes should not match because offset on the major hash is
    # only added from the top frame
    assert stack.minor != stack.major
    assert len(stack.frames) == 2


def test_stack_05():
    """test creating a Stack with 10 frames exceeding major depth"""
    frames = [
        BasicStackFrame(function="a", location="b", offset="c", stack_line=str(line))
        for line in range(10)
    ]
    stack = Stack(frames, major_depth=5)
    assert stack.minor is not None
    assert stack.major is not None
    # at this point the hashes should not match because offset on the major hash is
    # only added from the top frame
    assert stack.minor != stack.major
    assert len(stack.frames) == 10
    assert stack.major != Stack(frames, major_depth=4).major


def test_stack_06():
    """test Stack.from_text() - single Sanitizer stack"""
    input_txt = (
        "=================================================================\n"
        "==7854==ERROR: AddressSanitizer: SEGV on unknown address 0x000000000000"
        " (pc 0x7fcca620e0ee bp 0x7ffd946b2690 sp 0x7ffd946b25c0 T0)\n"
        "==7854==blah.\n"
        "==7854==Hint: blah.\n"
        "    #0 0x7fcca620e0ed in test::test::test(nsIWa*, nsICa*)"
        " /aa/bb/cc/dd/ee/ff/gg/asdf.cpp:5533:14\n"
        "    #1 0x7fcca3497201 in nsAs::TestTest(mz::dom::El*, int)"
        " /aa/bb/cc/dd/ee/ff/nsFilea.cpp:13733:3\n"
        "    #2 0x7fcca3495b9b in asdf::fasd()"
        " /aa/bb/cc/dd/ee/ff/base/nsFileb.cpp:11674:21\n"
        "    #3 0x7fcca02eb9d8 in nsAasd::PrNeEv(bool, bool*)"
        " /aa/bb/cc/dd/ee/xpcom/nsFilec.cpp:1396:14\n"
        "    #4 0x7fcca0307d40 in R_PNE(asd*, bool)"
        " /aa/bb/cc/dd/ee/xpcom/threads/asf.cpp:657:10\n"
        "    #5 0x2a780b25f65a  (<unknown module>)\n"
        "\n"
        "AddressSanitizer can not provide additional info.\n"
        "SUMMARY: AddressSanitizer: SEGV /aa/bb/cc/dd/ee/ff/asdf.cpp:5533:14"
        " in test::test::test(nsIWa*, nsICa*)\n"
        "==7854==ABORTING\n"
    )
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 6
    assert stack.minor != stack.major
    assert isinstance(stack.frames[0], SanitizerStackFrame)


def test_stack_07():
    """test Stack.from_text() - mixed frames modes"""
    input_txt = (
        ""
        "    #0 0x4d2cde in a_b_c /a/lib/info.c:392:12\n"
        "    #1 0x491e82 in main /a/b/d_e.c:128:8\n"
        "    #2 0x7f090384582f in __libc_start_main /build/a/../libc-start.c:291\n"
        "#0  0x0000000000400545 in gdb_frame ()\n"
        "    #3 0x41b228 in _start (bin_name+0x41b228)\n"
    )
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 4
    assert stack.minor != stack.major
    assert isinstance(stack.frames[0], SanitizerStackFrame)


def test_stack_08():
    """test Stack.from_text() - multiple stacks"""
    input_txt = (
        "=================================================================\n"
        "==5540==ERROR: AddressSanitizer: ...\n"
        "    #0 0x1badf00d in good::frame0(nsA*, nsB*) /aa/a.cpp:12:1\n"
        "    #1 0xdeadbeef in good::frame1(mz::d::EE*, int) /aa/a.cpp:12:1\n\n"
        "0x12876 is located 1024 bytes after 4096-byte region [0x12876,0x12876)\n"
        "freed by thread T5 here:\n"
        "    #0 0x0bad0bad in bad::frame0(nsA*, nsB*) /aa/a.cpp:12:1\n"
        "    #1 0x0bad0bad in bad::frame1(mz::d::EE*, int) /aa/a.cpp:12:1\n\n"
        "0x12876 is located 1024 bytes after 4096-byte region [0x12876,0x12876)\n"
        "previously allocated by thread T0 here:\n"
        "    #0 0x0bad0bad in bad::frame0(nsA*, nsB*) /aa/a.cpp:12:1\n"
        "    #1 0x0bad0bad in bad::frame1(mz::d::EE*, int) /aa/a.cpp:12:1\n\n"
    )
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 2
    assert stack.frames[0].function == "good::frame0"
    assert stack.frames[1].function == "good::frame1"
    assert stack.minor != stack.major
    assert isinstance(stack.frames[0], SanitizerStackFrame)


def test_stack_09():
    """test Stack.from_text() - empty string"""
    stack = Stack.from_text("")
    assert not stack.frames
    assert stack.minor is None
    assert stack.major is None


def test_stack_10():
    """test Stack.from_text() - Sanitizer trace with an unsymbolized lib"""
    input_txt = (
        ""
        "    #0 0x4c7702 in realloc asan/asan_malloc_linux.cc:107:3\n"
        "    #1 0x7f6d056ce7fc  (/lib/x86_64-linux-gnu/libdbus-1.so.3+0x2d7fc)\n"
        "    #2 0x7ffffffff  (<unknown module>)\n"
    )
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 3
    assert stack.frames[0].location == "asan_malloc_linux.cc"
    assert stack.frames[1].location == "libdbus-1.so.3"
    assert stack.frames[2].location == "<unknown module>"
    assert stack.minor != stack.major
    assert isinstance(stack.frames[0], SanitizerStackFrame)


def test_stack_11():
    """test Stack.from_text() - Sanitizer trace with an unsymbolized lib"""
    input_txt = (
        ""
        "    #0 0x90000223  (/usr/swr_a.so+0x231223)\n"
        "    #1 0x00000447  (/usr/as.so.1+0x42447)\n"
        "    #2 0x000098fc in fSasd /src/obj-firefox/dist/include/something.h:102:9\n"
        "    #3 0x000098fc in mz::as::asdf::SB() /src/Blah.cpp:655\n"
    )
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 4
    assert stack.frames[0].location == "swr_a.so"
    assert stack.frames[1].location == "as.so.1"
    assert stack.frames[2].function == "fSasd"
    assert stack.frames[3].function == "mz::as::asdf::SB"
    assert stack.minor != stack.major
    assert isinstance(stack.frames[0], SanitizerStackFrame)


def test_stack_12():
    """test Stack.from_text() - Valgrind trace"""
    input_txt = (
        ""
        "==4754== \n"
        "==4754== Use of uninitialised value of size 8\n"
        "==4754==    at 0x45C6C0: FooBar (decode.c:964)\n"
        "==4754==    by 0x462A20: main (test.cc:71)\n"
        "==4754==  Uninitialised value was created by a heap allocation\n"
        "==4754==    at 0x4C2AB80: malloc (in /usr/lib/test-linux.so)\n"
        "==4754==    by 0x459455: FooBar (decode.c:757)\n"
        "==4754==    by 0x462A20: main (test.cc:71)\n"
        "==4754== \n"
    )
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 5
    assert stack.frames[0].location == "decode.c"
    assert stack.frames[1].location == "test.cc"
    assert stack.frames[2].function == "malloc"
    assert stack.frames[3].function == "FooBar"
    assert stack.frames[4].function == "main"
    assert stack.minor != stack.major
    assert isinstance(stack.frames[0], ValgrindStackFrame)


def test_stack_13():
    """test Stack.from_text() - Rust trace"""
    input_txt = (
        "thread '<unnamed>' panicked at 'Invoking Servo_Element_IsDisplayContents"
        " on unstyled element', libcore/option.rs:917:5\n"
        "stack backtrace:\n"
        "   0:     0x7ff1c65e93d3 - std::sys::unix::backtrace::tracing::imp::unwind_bac"
        "ktrace::h09c1ee131a74b1c4\n"
        "                               at libstd/sys/unix/backtrace/tracing/gcc_s.rs:4"
        "9\n"
        "   1:     0x7ff1c65e81c4 - std::panicking::default_hook::{{closure}}::h945a649"
        "c9017832e\n"
        "                               at libstd/sys_common/backtrace.rs:71\n"
        "                               at libstd/sys_common/backtrace.rs:59\n"
        "                               at libstd/panicking.rs:380\n"
        "   2:     0x7ff1c65e7457 - std::panicking::default_hook::hcc534c2d30fbcda3\n"
        "                               at libstd/panicking.rs:396\n"
        "   3:     0x7ff1c65e6de7 - std::panicking::rust_panic_with_hook::h09a7a3a353dc"
        "2f38\n"
        "                               at libstd/panicking.rs:576\n"
        "   4:     0x7ff1c65e6c95 - std::panicking::begin_panic::h8327f16bde15df70\n"
        "                               at libstd/panicking.rs:537\n"
        "   5:     0x7ff1c65e6c29 - std::panicking::begin_panic_fmt::h42ff1d3740463d6\n"
        "                               at libstd/panicking.rs:521\n"
        "   6:     0x7ff1c65fa46a - core::panicking::panic_fmt::h0bd854df201d1baf\n"
        "                               at libstd/panicking.rs:497\n"
        "   7:     0x7ff1c65ffba8 - core::option::expect_failed::hfa0c8a51e07f7adc\n"
        "                               at libcore/option.rs:917\n"
        "   8:     0x7ff1c632d473 - Servo_Element_IsDisplayContents\n"
        "                               at /checkout/src/libcore/option.rs:302\n"
        "                               at servo/ports/geckolib/glue.rs:1086\n"
        "   9:     0x7f44064ee749 - _ZNK7nsFrame24DoGetParentComputedStyleEPPnsIFrame\n"
        "                                at /builds/worker/workspace/build/src/layout/g"
        "eneric/nsFrame.cpp:9893\n"
        "   10:     0x7f4406229749 - _ZN7mozilla14RestyleManager35DoReparentComputedSty"
        "leForFirstLineEP8nsIFrameRNS_13ServoStyleSetE\n"
        "                                at /builds/worker/workspace/build/src/layout/b"
        "ase/RestyleManager.cpp:3407\n"
        "   11:     0x7f440622a0a5 - _ZN7mozilla14RestyleManager24ReparentFrameDescenda"
        "ntsEP8nsIFrameS2_RNS_13ServoStyleSetE\n"
        "                                at /builds/worker/workspace/build/src/layout/b"
        "ase/RestyleManager.cpp:3538\n"
    )
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 12
    assert (
        stack.frames[0].function
        == "std::sys::unix::backtrace::tracing::imp::unwind_backtrace"
    )
    assert stack.frames[8].function == "Servo_Element_IsDisplayContents"
    assert stack.minor != stack.major
    assert isinstance(stack.frames[0], RustStackFrame)


def test_stack_14():
    """test Stack.height_limit"""
    frames = [
        BasicStackFrame(
            function=str(num), location="b", offset="c", stack_line=str(num)
        )
        for num in range(10)
    ]
    stack = Stack(frames, major_depth=3)
    assert stack.height_limit == 0
    no_lim_minor = stack.minor
    assert no_lim_minor is not None
    no_lim_major = stack.major
    assert no_lim_major is not None
    # set height limit and check hash recalculations
    stack.height_limit = 5
    assert stack.height_limit == 5
    assert stack.minor is not None
    assert no_lim_minor != stack.minor
    assert stack.major is not None
    assert no_lim_major != stack.major
    # remove height limit and check hash recalculations
    stack.height_limit = 0
    assert stack.height_limit == 0
    assert no_lim_minor == stack.minor
    assert no_lim_major == stack.major


def test_stack_15():
    """test Stack.from_text() - trace missing #0"""
    stack = Stack.from_text("    #1 0x000098fc in frame1() test/a.cpp:655\n")
    assert len(stack.frames) == 1
    assert stack.frames[0].location == "a.cpp"
    assert stack.frames[0].function == "frame1"
    assert isinstance(stack.frames[0], SanitizerStackFrame)


def test_stack_16():
    """test Stack.from_text() - mixed traces"""
    stack = Stack.from_text(
        ""
        "    #0 0x0001230 in stack_1a() test/a.cpp:12\n"
        "    #2 0x0001231 in stack_2a() test/a.cpp:45\n"
        "    #1 0x0001232 in stack_1b() test/a.cpp:23\n"
        "    #2 0x0001233 in stack_1c() test/a.cpp:34\n"
    )
    assert len(stack.frames) == 1
    assert stack.frames[0].function == "stack_1a"
    assert isinstance(stack.frames[0], SanitizerStackFrame)


def test_stack_17():
    """test Stack.from_text() - contains ignored frames"""
    st_01 = (
        "#0 0x10 in MOZ_Crash /a.h:281:3\n"
        "#1 0x11 in std::panicking::ignored::hd80c17bcc51bbfda /l.rs:96:9\n"
        "#2 0x11 in std::panicking::ignored::hd80c17bcc51bbfda /l.rs:96:9\n"
        "#3 0x11 in foo_a /l.rs:1:9\n"
        "#4 0x11 in foo_b1111 /l.rs:2:9\n"
        "#5 0x11 in foo_c1111 /l.rs:3:9\n"
    )
    stack01 = Stack.from_text(st_01, major_depth=3)
    assert len(stack01.frames) == 6

    st_02 = (
        "#0 0x10 in MOZ_Crash /a.h:281:3\n"
        "#1 0x11 in std::panicking::ignored::hd80c17bcc51bbfda /l.rs:96:9\n"
        "#2 0x11 in std::panicking::ignored::hd80c17bcc51bbfda /l.rs:96:9\n"
        "#3 0x11 in foo_a /l.rs:1:9\n"
        "#4 0x11 in foo_b2222 /l.rs:2:9\n"
        "#5 0x11 in foo_c2222 /l.rs:3:9\n"
    )
    stack02 = Stack.from_text(st_02, major_depth=3)
    assert len(stack02.frames) == 6
    assert stack01.minor != stack02.minor
    assert stack01.major != stack02.major
    assert isinstance(stack01.frames[0], SanitizerStackFrame)


def test_stack_18():
    """test Stack.from_text() - multiple minidump stackwalk traces"""
    stack = Stack.from_text(
        "CPU|x86|GenuineIntel family 6 model 85 stepping 4|8\n"
        "Crash|EXCEPTION_BREAKPOINT|0x67a30091|44\n"
        "42|0|xul.dll|foo_a()|s3:g-g-sources:e/a.cpp:|14302|0x3f1\n"
        "42|1|xul.dll|foo_b()|s3:g-g-sources:e/a.cpp:|14302|0x3f1\n"
        "42|2|xul.dll|foo_c()|s3:g-g-sources:e/a.cpp:|14302|0x3f1\n"
        "43|0|xul.dll|bar_a()|s3:g-g-sources:e/a.cpp:|14302|0x3f1\n"
        "43|1|xul.dll|bar_b()|s3:g-g-sources:e/a.cpp:|14302|0x3f1\n"
        "43|2|xul.dll|bar_c()|s3:g-g-sources:e/a.cpp:|14302|0x3f1\n"
        "44|0|xul.dll|good_a()|s3:g-g-sources:e/a.cpp:|14302|0x3f1\n"
        "44|1|xul.dll|good_a()|s3:g-g-sources:e/a.cpp:|14302|0x3f1\n"
        "44|2|xul.dll|good_c()|s3:g-g-sources:e/a.cpp:|14302|0x3f1\n"
    )
    assert len(stack.frames) == 3
    assert stack.frames[0].function == "good_a()"
    assert isinstance(stack.frames[0], MinidumpStackFrame)


def test_stackframe_01():
    """test creating an empty generic BasicStackFrame"""
    stack = BasicStackFrame()
    assert not str(stack)


@mark.parametrize(
    "frame_class",
    (
        GdbStackFrame,
        MinidumpStackFrame,
        RrStackFrame,
        RustStackFrame,
        SanitizerStackFrame,
        ThreadSanitizerStackFrame,
        ValgrindStackFrame,
    ),
)
@mark.parametrize(
    "input_line",
    (
        "#0      ",
        " #0 ",
        "#0#0#0#0#0#0#0#0",
        "#a",
        "",
        "###",
        "123",
        "test()",
        "|||",
        "||||||",
        "a|b|c|d|e|f|g",
        "==123==",
        "==1== by 0x0: a ()",
        "rr(foo",
        "==1==    at 0x0: ??? (:)",
    ),
)
def test_stackframe_02(frame_class, input_line):
    """test StackFrame.from_line() - junk"""
    assert frame_class.from_line(input_line) is None


def test_sanitizer_stackframe_01():
    """test SanitizerStackFrame.from_line() - with symbols"""
    frame = SanitizerStackFrame.from_line(
        "    #1 0x7f00dad60565 in Abort(char const*) /blah/base/nsDebugImpl.cpp:472"
    )
    assert frame
    assert frame.stack_line == "1"
    assert frame.function == "Abort"
    assert frame.location == "nsDebugImpl.cpp"
    assert frame.offset == "472"


def test_sanitizer_stackframe_02():
    """test SanitizerStackFrame.from_line() - with symbols"""
    frame = SanitizerStackFrame.from_line(
        "    #36 0x48a6e4 in main /app/nsBrowserApp.cpp:399:11"
    )
    assert frame
    assert frame.stack_line == "36"
    assert frame.function == "main"
    assert frame.location == "nsBrowserApp.cpp"
    assert frame.offset == "399"


def test_sanitizer_stackframe_03():
    """test SanitizerStackFrame.from_line() - without symbols"""
    frame = SanitizerStackFrame.from_line(
        "    #1 0x7f00ecc1b33f (/lib/x86_64-linux-gnu/libpthread.so.0+0x1033f)"
    )

    assert frame
    assert frame.stack_line == "1"
    assert frame.function is None
    assert frame.location == "libpthread.so.0"
    assert frame.offset == "0x1033f"


def test_sanitizer_stackframe_04():
    """test SanitizerStackFrame.from_line() - with symbols"""
    frame = SanitizerStackFrame.from_line(
        "    #25 0x7f0155526181 in start_thread (/l/libpthread.so.0+0x8181)"
    )
    assert frame
    assert frame.stack_line == "25"
    assert frame.function == "start_thread"
    assert frame.location == "libpthread.so.0"
    assert frame.offset == "0x8181"


def test_sanitizer_stackframe_05():
    """test SanitizerStackFrame.from_line() - angle brackets"""
    frame = SanitizerStackFrame.from_line(
        "    #123 0x7f30afea9148 in Call<nsBlah *> /a/b.cpp:356:50"
    )
    assert frame
    assert frame.stack_line == "123"
    assert frame.function == "Call"
    assert frame.location == "b.cpp"
    assert frame.offset == "356"


def test_sanitizer_stackframe_06():
    """test SanitizerStackFrame.from_line() - useless frame"""
    frame = SanitizerStackFrame.from_line("    #2 0x7ffffffff  (<unknown module>)")
    assert frame
    assert frame.stack_line == "2"
    assert frame.function is None
    assert frame.location == "<unknown module>"
    assert frame.offset is None


def test_sanitizer_stackframe_07():
    """test SanitizerStackFrame.from_line() - missing a function"""
    frame = SanitizerStackFrame.from_line(
        "    #0 0x7f0d571e04bd  /a/glibc-2.23/../syscall-template.S:84"
    )
    assert frame
    assert frame.stack_line == "0"
    assert frame.function is None
    assert frame.location == "syscall-template.S"
    assert frame.offset == "84"


def test_sanitizer_stackframe_08():
    """test SanitizerStackFrame.from_line() - lots of spaces"""
    frame = SanitizerStackFrame.from_line(
        "    #0 0x48a6e4 in Call<a *> /test path/file name.c:1:2"
    )
    assert frame
    assert frame.stack_line == "0"
    assert frame.function == "Call"
    assert frame.location == "file name.c"
    assert frame.offset == "1"


def test_sanitizer_stackframe_09():
    """test SanitizerStackFrame.from_line() - filename missing path"""
    frame = SanitizerStackFrame.from_line("    #0 0x0000123 in func a.cpp:12")
    assert frame
    assert frame.stack_line == "0"
    assert frame.function == "func"
    assert frame.location == "a.cpp"
    assert frame.offset == "12"


def test_sanitizer_stackframe_10():
    """test SanitizerStackFrame.from_line() - with build id"""
    frame = SanitizerStackFrame.from_line(
        "    #0 0x7f76d25b7fc0  (/usr/lib/x86_64-linux-gnu/dri/swrast_dri.so+0x704fc0) "
        "(BuildId: d04a40e4062a8d444ff6f23d4fe768215b2e32c7)"
    )
    assert frame
    assert frame.stack_line == "0"
    assert frame.function is None
    assert frame.location == "swrast_dri.so"
    assert frame.offset == "0x704fc0"


def test_gdb_stackframe_01():
    """test GdbStackFrame.from_line() - with symbols"""
    frame = GdbStackFrame.from_line(
        "#0  __memmove_ssse3_back () at ../d/x86_64/a/memcpy-ssse3-back.S:1654"
    )
    assert frame
    assert frame.stack_line == "0"
    assert frame.function == "__memmove_ssse3_back"
    assert frame.location == "memcpy-ssse3-back.S"
    assert frame.offset == "1654"


def test_gdb_stackframe_02():
    """test GdbStackFrame.from_line() - with symbols, missing line numbers"""
    frame = GdbStackFrame.from_line("#2  0x0000000000400545 in main ()")
    assert frame
    assert frame.stack_line == "2"
    assert frame.function == "main"
    assert frame.location is None
    assert frame.offset is None


def test_gdb_stackframe_03():
    """test GdbStackFrame.from_line() - with symbols"""
    frame = GdbStackFrame.from_line("#3  0x0000000000400545 in main () at test.c:5")
    assert frame
    assert frame.stack_line == "3"
    assert frame.function == "main"
    assert frame.location == "test.c"
    assert frame.offset == "5"


def test_gdb_stackframe_04():
    """test GdbStackFrame.from_line() - unknown address"""
    frame = GdbStackFrame.from_line("#0  0x00000000 in ?? ()")
    assert frame
    assert frame.stack_line == "0"
    assert frame.function == "??"
    assert frame.location is None
    assert frame.offset is None


def test_gdb_stackframe_05():
    """test GdbStackFrame.from_line() - missing line number"""
    frame = GdbStackFrame.from_line("#3  0x400545 in main () at test.c")
    assert frame
    assert frame.stack_line == "3"
    assert frame.function == "main"
    assert frame.location == "test.c"
    assert frame.offset is None


def test_minidump_stackframe_01():
    """test MinidumpStackFrame.from_line() - with symbols"""
    frame = MinidumpStackFrame.from_line(
        "0|2|libtest|main|hg:c.a.org/m-c:a/b/file.cpp:5bf50|114|0x3a"
    )
    assert frame
    assert frame.stack_line == "2"
    assert frame.function == "main"
    assert frame.location == "file.cpp"
    assert frame.offset == "114"


def test_minidump_stackframe_02():
    """test MinidumpStackFrame.from_line() - without symbols"""
    frame = MinidumpStackFrame.from_line("9|42|libpthread-2.26.so||||0x10588")
    assert frame
    assert frame.stack_line == "42"
    assert frame.function is None
    assert frame.location == "libpthread-2.26.so"
    assert frame.offset == "0x10588"


def test_minidump_stackframe_03():
    """test MinidumpStackFrame.from_line() - without hg repo info"""
    frame = MinidumpStackFrame.from_line(
        "0|49|libxul.so|foo|/usr/x86_64-linux-gnu/test.h|85|0x5"
    )
    assert frame
    assert frame.stack_line == "49"
    assert frame.function == "foo"
    assert frame.location == "/usr/x86_64-linux-gnu/test.h"
    assert frame.offset == "85"


def test_minidump_stackframe_04():
    """test MinidumpStackFrame.from_line() - with s3 repo info"""
    frame = MinidumpStackFrame.from_line(
        "42|0|xul.dll|foo_a() const|s3:g-g-sources:e/a.cpp:|14302|0x3f1"
    )
    assert frame
    assert frame.stack_line == "0"
    assert frame.function == "foo_a() const"
    assert frame.location == "a.cpp"
    assert frame.offset == "14302"


def test_tsan_stackframe_01():
    """test ThreadSanitizerStackFrame.from_line() - symbolized"""
    frame = ThreadSanitizerStackFrame.from_line("    #0 main race.c:10 (exe+0xa3b4)")
    assert frame
    assert frame.stack_line == "0"
    assert frame.function == "main"
    assert frame.location == "race.c"
    assert frame.offset == "10"


def test_tsan_stackframe_02():
    """test ThreadSanitizerStackFrame.from_line() - symbolized"""
    frame = ThreadSanitizerStackFrame.from_line(
        "    #1 test1 test2 /a b/c.h:51:10 (libxul.so+0x18c9873)"
    )
    assert frame
    assert frame.stack_line == "1"
    assert frame.function == "test1"
    assert frame.location == "c.h"
    assert frame.offset == "51"


def test_tsan_stackframe_03():
    """test ThreadSanitizerStackFrame.from_line() - unsymbolized"""
    frame = ThreadSanitizerStackFrame.from_line("    #2 <null> <null> (0xbad)")
    assert frame
    assert frame.stack_line == "2"
    assert frame.function is None
    assert frame.location is None
    assert frame.offset == "0xbad"


def test_tsan_stackframe_04():
    """test ThreadSanitizerStackFrame.from_line() - missing file"""
    frame = ThreadSanitizerStackFrame.from_line("    #0 func <null> (mod+0x123ac)")
    assert frame
    assert frame.stack_line == "0"
    assert frame.function == "func"
    assert frame.location == "mod"
    assert frame.offset == "0x123ac"


def test_valgrind_stackframe_01():
    """test ValgrindStackFrame.from_line()"""
    frame = ValgrindStackFrame.from_line(
        "==4754==    at 0x45C6C0: FuncName (decode.c:123)"
    )
    assert frame
    assert frame.stack_line is None
    assert frame.function == "FuncName"
    assert frame.location == "decode.c"
    assert frame.offset == "123"


def test_valgrind_stackframe_02():
    """test ValgrindStackFrame.from_line()"""
    frame = ValgrindStackFrame.from_line("==4754==    by 0x462A20: main (foo.cc:71)")
    assert frame
    assert frame.stack_line is None
    assert frame.function == "main"
    assert frame.location == "foo.cc"
    assert frame.offset == "71"


def test_valgrind_stackframe_03():
    """test ValgrindStackFrame.from_line()"""
    frame = ValgrindStackFrame.from_line(
        "==4754==    at 0x4C2AB80: malloc (in /usr/lib/blah-linux.so)"
    )
    assert frame
    assert frame.stack_line is None
    assert frame.function == "malloc"
    assert frame.location == "blah-linux.so"
    assert frame.offset is None


def test_valgrind_stackframe_04():
    """test ValgrindStackFrame.from_line()"""
    frame = ValgrindStackFrame.from_line(
        "==2342==    by 0x4E3E71: (anon ns)::test(b2::a&, int) (main.cpp:49)"
    )
    assert frame
    assert frame.stack_line is None
    assert frame.function == "(anon ns)::test(b2::a&, int)"
    assert frame.location == "main.cpp"
    assert frame.offset == "49"


def test_valgrind_stackframe_05():
    """test ValgrindStackFrame.from_line()"""
    frame = ValgrindStackFrame.from_line(
        "==2342==    at 0xF00D: Foo::Foo(char *, int, bool) (File.h:37)"
    )
    assert frame
    assert frame.stack_line is None
    assert frame.function == "Foo::Foo(char *, int, bool)"
    assert frame.location == "File.h"
    assert frame.offset == "37"


def test_valgrind_stackframe_06():
    """test ValgrindStackFrame.from_line()"""
    frame = ValgrindStackFrame.from_line("==4754==    at 0x4C2AB80: ??? (in /bin/a)")
    assert frame
    assert frame.stack_line is None
    assert frame.function == "???"
    assert frame.location == "a"
    assert frame.offset is None


def test_rr_stackframe_01():
    """test RrStackFrame.from_line()"""
    frame = RrStackFrame.from_line("rr(main+0x244)[0x450b74]")
    assert frame
    assert frame.stack_line is None
    assert frame.function is None
    assert frame.location == "main"
    assert frame.offset == "0x244"


def test_rust_stackframe_01():
    """test RustStackFrame.from_line()"""
    frame = RustStackFrame.from_line("  53:    0x7ff1d7e4982f - __libc_start_main")
    assert frame
    assert frame.stack_line == "53"
    assert frame.function == "__libc_start_main"
    assert frame.location is None
    assert frame.offset is None


def test_rust_stackframe_02():
    """test RustStackFrame.from_line()"""
    frame = RustStackFrame.from_line(
        "  4:    0x10b715a5b - unwind::begin_unwind_fmt::h227376fe1e021a36n3d"
    )
    assert frame
    assert frame.stack_line == "4"
    assert frame.location is None
    assert frame.function == "unwind::begin_unwind_fmt"
    assert frame.offset is None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import json
from dataclasses import dataclass, field
from itertools import chain, product
from logging import getLogger
from os.path import normpath, split
from pathlib import Path
from shutil import copyfile, copytree, move, rmtree
from tempfile import NamedTemporaryFile, mkdtemp
from time import time
from typing import Any, Generator, Iterator, cast

from .utils import grz_tmp, package_version

__all__ = ("TestCase", "TestCaseLoadFailure", "TestFileExists")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


LOG = getLogger(__name__)
GRZ_VERSION = package_version("grizzly-framework")
TEST_INFO = "test_info.json"


class TestCaseLoadFailure(Exception):
    """Raised when loading a TestCase fails"""


class TestFileExists(Exception):
    """Raised when adding a TestFile to a TestCase that has an existing
    TestFile with the same name"""


@dataclass(eq=False)
class TestFileMap:
    optional: dict[str, Path] = field(default_factory=dict)
    required: dict[str, Path] = field(default_factory=dict)


class TestCase:
    __slots__ = (
        "adapter_name",
        "assets",
        "assets_path",
        "duration",
        "entry_point",
        "env_vars",
        "hang",
        "https",
        "input_fname",
        "timestamp",
        "version",
        "_files",
        "_in_place",
        "_root",
    )

    def __init__(
        self,
        entry_point: str,
        adapter_name: str,
        data_path: Path | None = None,
        input_fname: str | None = None,
        timestamp: float | None = None,
    ) -> None:
        assert entry_point
        self.adapter_name = adapter_name
        self.assets: dict[str, str] = {}
        self.assets_path: Path | None = None
        self.duration: float | None = None
        self.env_vars: dict[str, str] = {}
        self.hang = False
        self.https = False
        self.input_fname = input_fname  # file that was used to create the test case
        self.entry_point = self.sanitize_path(entry_point)
        self.timestamp = time() if timestamp is None else timestamp
        self.version = GRZ_VERSION
        self._files = TestFileMap()
        if data_path is not None:
            self._root = data_path
            self._in_place = True
        else:
            self._root = Path(mkdtemp(prefix="testcase_", dir=grz_tmp("storage")))
            self._in_place = False

    def __contains__(self, item: str) -> bool:
        """Scan TestCase contents for a URL.

        Args:
            item: URL.

        Returns:
            URL found in contents.
        """
        return item in self._files.required or item in self._files.optional

    def __getitem__(self, key: str) -> Path:
        """Lookup file path by URL.

        Args:
            key: URL.

        Returns:
            Test file.
        """
        if key in self._files.required:
            return self._files.required[key]
        return self._files.optional[key]

    def __enter__(self) -> TestCase:
        return self

    def __exit__(self, *exc: object) -> None:
        self.cleanup()

    def __iter__(self) -> Iterator[str]:
        """TestCase contents.

        Args:
            None

        Yields:
            URLs of contents.
        """
        yield from self._files.required
        yield from self._files.optional

    def __len__(self) -> int:
        """Number files in TestCase.

        Args:
            None

        Returns:
            Number files in TestCase.
        """
        return len(self._files.optional) + len(self._files.required)

    def add_from_bytes(
        self, data: bytes, file_name: str, required: bool = False
    ) -> None:
        """Create a file and add it to the TestCase.

        Args:
            data: Data to write to file.
            file_name: Used as file path on disk and URI. Relative to wwwroot.
            required: Indicates whether the file must be served.

        Returns:
            None
        """
        with NamedTemporaryFile(
            delete=False, dir=grz_tmp("storage"), mode="w+b"
        ) as in_fp:
            in_fp.write(data)
            data_file = Path(in_fp.name)

        try:
            self.add_from_file(
                data_file, file_name=file_name, required=required, copy=False
            )
        finally:
            # the temporary file should have been moved to the data path of the TestCase
            # unless an exception occurred so remove it if needed
            data_file.unlink(missing_ok=True)

    def add_from_file(
        self,
        src_file: str | Path,
        file_name: str | None = None,
        required: bool = False,
        copy: bool = False,
    ) -> None:
        """Add a file to the TestCase. Copy or move an existing file if needed.

        Args:
            src_file: Path to existing file to use.
            file_name: Used as file path on disk and URI. Relative to wwwroot.
                       If file_name is not given the name of the src_file
                       will be used.
            required: Indicates whether the file must be served. Typically this
                      is only used for the entry point.
            copy: Copy existing file data. Existing data is moved by default.

        Returns:
            None
        """
        src_file = Path(src_file)
        if file_name is None:
            url_path = self.sanitize_path(src_file.name)
        else:
            url_path = self.sanitize_path(file_name)
        if url_path in self:
            raise TestFileExists(f"{url_path!r} exists in test")

        dst_file = self._root / url_path
        # don't move/copy data is already in place
        if src_file.resolve() != dst_file.resolve():
            assert not self._in_place
            dst_file.parent.mkdir(parents=True, exist_ok=True)
            if copy:
                copyfile(src_file, dst_file)
            else:
                # Python 3.9+: move() accepts a path-like object for both src and dst
                move(str(src_file.resolve()), str(dst_file.resolve()))

        # entry_point is always 'required'
        if required or url_path == self.entry_point:
            self._files.required[url_path] = dst_file
        else:
            self._files.optional[url_path] = dst_file

    def cleanup(self) -> None:
        """Remove all the test files.

        Args:
            None

        Returns:
            None
        """
        if not self._in_place:
            rmtree(self._root, ignore_errors=True)

    def clear_optional(self) -> None:
        """Clear optional files. This does not remove data from the file system.

        Args:
            None

        Returns:
            None
        """
        self._files.optional.clear()

    def clone(self) -> TestCase:
        """Make a copy of the TestCase.

        Args:
            None

        Returns:
            A copy of the TestCase instance.
        """
        result = type(self)(
            self.entry_point,
            self.adapter_name,
            input_fname=self.input_fname,
            timestamp=self.timestamp,
        )
        result.assets = dict(self.assets)
        if result.assets:
            assert self.assets_path
            try:
                # copy asset data from test case
                result.assets_path = result.root / self.assets_path.relative_to(
                    self.root
                )
                copytree(self.assets_path, result.assets_path)
            except ValueError:
                # asset data is not part of the test case
                result.assets_path = self.assets_path
        result.duration = self.duration
        result.env_vars = dict(self.env_vars)
        result.hang = self.hang
        result.https = self.https
        # copy test data files
        for (file_name, data_file), required in chain(
            product(self._files.required.items(), [True]),
            product(self._files.optional.items(), [False]),
        ):
            result.add_from_file(
                data_file, file_name=file_name, required=required, copy=True
            )
        return result

    @property
    def data_size(self) -> int:
        """Total amount of data used (in bytes) by the files in the TestCase.

        Args:
            None

        Returns:
            Total size of the TestCase in bytes.
        """
        total = 0
        for data_file in chain(
            self._files.required.values(),
            self._files.optional.values(),
        ):
            total += data_file.stat().st_size
        return total

    def dump(self, dst_path: Path, include_details: bool = False) -> None:
        """Write all the test case data to the filesystem.

        Args:
            dst_path: Path to directory to output data.
            include_details: Output test info file.

        Returns:
            None
        """
        # save test files to dst_path
        for src_url, src_file in chain(
            self._files.required.items(),
            self._files.optional.items(),
        ):
            dst_file = dst_path / src_url
            dst_file.parent.mkdir(parents=True, exist_ok=True)
            copyfile(src_file, dst_file)
        # save test case files and meta data including:
        # adapter used, input file, environment info and files
        if include_details:
            info = {
                "https": self.https,
                "target": self.entry_point,
            }
            if self.adapter_name:
                info["adapter"] = self.adapter_name
            if self.duration:
                info["duration"] = self.duration
            if self.env_vars:
                info["env"] = self.env_vars
            if self.hang:
                info["hang"] = self.hang
            if self.input_fname:
                info["input"] = Path(self.input_fname).name
            if self.timestamp:
                info["timestamp"] = self.timestamp
            if self.version:
                info["version"] = self.version
            # save target assets and update meta data
            if self.assets:
                assert isinstance(self.assets_path, Path)
                info["assets"] = self.assets
                info["assets_path"] = "_assets_"
                copytree(self.assets_path, dst_path / cast(str, info["assets_path"]))
            with (dst_path / TEST_INFO).open("w") as out_fp:
                json.dump(info, out_fp, indent=2, sort_keys=True)

    @staticmethod
    def _find_entry_point(path: Path) -> Path:
        """Locate potential entry point.

        Args:
            path: Directory to scan.

        Returns:
            Entry point.
        """
        entry_point = None
        for entry in path.iterdir():
            if entry.suffix.lower() in (".htm", ".html"):
                if entry_point is not None:
                    raise TestCaseLoadFailure("Ambiguous entry point")
                entry_point = entry
        if entry_point is None:
            raise TestCaseLoadFailure("Could not determine entry point")
        return entry_point

    @classmethod
    def load(
        cls, path: Path, entry_point: Path | None = None, catalog: bool = False
    ) -> TestCase:
        """Load a TestCase.

        Args:
            path: Path can be:
                - A single file to be used as a test case.
                - A directory containing the test case data.
            entry_point: File to use as entry point.
            catalog: Scan contents of TestCase.root and track files.
                Untracked files will be missed when using clone() or dump().
                Only the entry point will be marked as 'required'.

        Returns:
            A TestCase.
        """
        assert isinstance(path, Path)
        # load test case info
        entry_point, info = cls.load_meta(path, entry_point=entry_point)
        # create test case
        test = cls(
            entry_point.relative_to(entry_point.parent).as_posix(),
            info.get("adapter", None),
            data_path=entry_point.parent,
            input_fname=info.get("input", None),
            timestamp=info.get("timestamp", 0),
        )
        test.assets = info.get("assets", {})
        test.duration = info.get("duration", None)
        test.env_vars = info.get("env", {})
        test.hang = info.get("hang", False)
        test.https = info.get("https", False)
        test.version = info.get("version", None)
        # sanity check assets data
        assert isinstance(test.assets, dict)
        for name, value in test.assets.items():
            if not isinstance(name, str) or not isinstance(value, str):
                raise TestCaseLoadFailure("'assets' contains invalid entry")
        if test.assets:
            assets_path = info.get("assets_path", None)
            if not assets_path or not (test.root / assets_path).is_dir():
                LOG.warning("Could not find assets in test case")
                test.assets = {}
            else:
                test.assets_path = test.root / assets_path
        # sanity check environment variable data
        assert isinstance(test.env_vars, dict)
        for name, value in test.env_vars.items():
            if not isinstance(name, str) or not isinstance(value, str):
                raise TestCaseLoadFailure("'env' contains invalid entry")
        # add contents of directory to test case 'contents' (excluding assets)
        # data is not copied/moved because it is already in place
        if catalog and path.is_dir():
            # NOTE: only entry point will be marked as 'required'
            for entry in test.root.rglob("*"):
                if (
                    not entry.is_dir()
                    and test.assets_path not in entry.parents
                    and entry.name != TEST_INFO
                ):
                    test.add_from_file(
                        entry, file_name=entry.relative_to(test.root).as_posix()
                    )
        else:
            # add entry point
            test.add_from_file(entry_point, required=True)
        return test

    @classmethod
    def load_meta(
        cls, path: Path, entry_point: Path | None = None
    ) -> tuple[Path, dict[str, Any]]:
        """Process and sanitize TestCase meta data.

        Args:
            path: Directory containing test info file.
            entry_point: See TestCase.load().

        Returns:
            Test case entry point and loaded test info.
        """
        # load test case info if available
        if path.is_dir():
            try:
                info = cls.read_info(path)
            except TestCaseLoadFailure as exc:
                LOG.info(exc)
                info = {}
            if entry_point is not None:
                info["target"] = entry_point.name
            elif info:
                entry_point = path / info["target"]
                assert entry_point
            else:
                # attempt to determine entry point
                entry_point = cls._find_entry_point(path)
            if path not in entry_point.parents:
                raise TestCaseLoadFailure("Entry point must be in root of given path")
        else:
            # single file test case
            entry_point = path
            info = {}

        if not entry_point.exists():
            raise TestCaseLoadFailure(f"Missing or invalid TestCase '{path}'")

        return (entry_point, info)

    @property
    def optional(self) -> Generator[str, None, None]:
        """Get file paths of optional files.

        Args:
            None

        Yields:
            File path of each optional file.
        """
        yield from self._files.optional

    @staticmethod
    def read_info(path: Path) -> dict[str, Any]:
        """Attempt to load test info.

        Args:
            path: Directory containing test info file.

        Returns:
            Test info.
        """
        try:
            with (path / TEST_INFO).open("r") as in_fp:
                info = json.load(in_fp)
        except FileNotFoundError:
            info = None
        except ValueError:
            raise TestCaseLoadFailure(f"Invalid '{TEST_INFO}'") from None
        if info is not None and not isinstance(info.get("target"), str):
            raise TestCaseLoadFailure(f"Invalid 'target' entry in '{TEST_INFO}'")
        return info or {}

    @property
    def required(self) -> Generator[str, None, None]:
        """Get file paths of required files.

        Args:
            None

        Yields:
            File path of each file.
        """
        yield from self._files.required

    @property
    def root(self) -> Path:
        """Location test data is stored on disk. This is intended to be used as wwwroot.

        Args:
            None

        Returns:
            Directory containing test case files.
        """
        return self._root

    @staticmethod
    def sanitize_path(path: str) -> str:
        """Sanitize given path for use as a URI path.

        Args:
            path: Path to sanitize. Must be relative to wwwroot.

        Returns:
            Sanitized path.
        """
        assert isinstance(path, str)
        # check for missing filename or path containing drive letter (Windows)
        if split(path)[-1] in ("", ".", "..") or ":" in path:
            raise ValueError(f"invalid path {path!r}")
        # normalize path
        path = normpath(path).replace("\\", "/")
        # check normalized path does not resolve to location outside of '.'
        if path.startswith("../"):
            raise ValueError(f"invalid path {path!r}")
        return path.lstrip("/")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Sapphire HTTP server worker
"""
from __future__ import annotations

from logging import getLogger
from re import compile as re_compile
from select import select
from socket import SHUT_RDWR, socket
from socket import timeout as sock_timeout  # Py3.10 socket.timeout => TimeoutError
from sys import exc_info
from threading import Thread, ThreadError, active_count
from time import sleep
from typing import TYPE_CHECKING
from urllib.parse import ParseResult, quote, unquote, urlparse

from .server_map import DynamicResource, FileResource, RedirectResource

if TYPE_CHECKING:
    from .job import Job

# TODO: urlparse -> urlsplit


__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


class Request:
    REQ_PATTERN = re_compile(rb"^(?P<method>\w+)\s(?P<url>\S+)\sHTTP/1")

    __slots__ = ("method", "url")

    def __init__(self, method: str, url: ParseResult) -> None:
        self.method = method
        self.url = url

    @classmethod
    def parse(cls, raw_data: bytes) -> Request | None:
        assert isinstance(raw_data, bytes)
        req_match = cls.REQ_PATTERN.match(raw_data)
        if not req_match:
            LOG.debug("request failed to match regex")
            return None

        # TODO: parse headers if needed

        try:
            url_str = req_match.group("url").decode("ascii", errors="replace")
            # unquote() accepts str | bytes as of Python 3.9
            url = urlparse(unquote(url_str))
        except ValueError as exc:
            msg = str(exc)
            if (
                "contains invalid characters under NFKC normalization" not in msg
                and "Invalid IPv6 URL" not in msg
                and "does not appear to be an IPv4 or IPv6 address" not in msg
            ):
                LOG.error("Failed to parse URL: %r", url_str)
                raise
            LOG.debug("failed to parse url from request")
            return None

        return cls(req_match.group("method").decode("ascii", errors="replace"), url)


class WorkerError(Exception):
    """Raised by Worker"""


class Worker:
    DEFAULT_REQUEST_LIMIT = 0x1000  # 4KB
    DEFAULT_TX_SIZE = 0x10000  # 64KB

    __slots__ = ("_conn", "_thread")

    def __init__(self, conn: socket, thread: Thread) -> None:
        self._conn = conn
        self._thread: Thread | None = thread

    @staticmethod
    def _200_header(c_length: int, c_type: str) -> bytes:
        assert c_type is not None
        data = (
            "HTTP/1.1 200 OK\r\n"
            "Cache-Control: max-age=0, no-cache\r\n"
            f"Content-Length: {c_length}\r\n"
            f"Content-Type: {c_type}\r\n"
            "Connection: close\r\n\r\n"
        )
        return data.encode(encoding="ascii")

    @staticmethod
    def _307_redirect(redirect_to: str) -> bytes:
        data = (
            "HTTP/1.1 307 Temporary Redirect\r\n"
            f"Location: {redirect_to}\r\n"
            "Connection: close\r\n\r\n"
        )
        return data.encode(encoding="ascii")

    @staticmethod
    def _4xx_page(code: int, hdr_msg: str, close: int = -1) -> bytes:
        if close < 0:
            content = f"<h3>{code}!</h3>"
        else:
            content = (
                "<script>\n"
                "window.onload = () => "
                f"{{ window.setTimeout(window.close, {close * 1000}) }}\n"
                "</script>\n"
                '<body style="background-color:#ffffe0">\n'
                f"<h3>{code}! - Calling window.close() in {close} seconds</h3>\n"
                "</body>\n"
            )
        data = (
            f"HTTP/1.1 {code} {hdr_msg}\r\n"
            f"Content-Length: {len(content)}\r\n"
            "Content-Type: text/html\r\n"
            f"Connection: close\r\n\r\n{content}"
        )
        return data.encode(encoding="ascii")

    def close(self) -> None:
        # workers that are no longer running will have had close() called
        if self.is_alive():
            # shutdown socket to avoid hang
            LOG.debug("closing socket while thread is running!")
            try:
                self._conn.shutdown(SHUT_RDWR)
            except OSError as exc:
                LOG.debug("close - shutdown(): %s", exc)
            self._conn.close()

    def is_alive(self) -> bool:
        return self._thread is not None and self._thread.is_alive()

    @classmethod
    def handle_request(cls, conn: socket, serv_job: Job) -> None:
        finish_job = False  # call finish() on return
        try:
            # socket operations should not block forever
            assert conn.gettimeout() is not None
            # receive incoming request data
            raw_request = conn.recv(cls.DEFAULT_REQUEST_LIMIT)
            if not raw_request:
                LOG.debug("request was empty")
                serv_job.accepting.set()
                return
            request = Request.parse(raw_request)
            # handle bad request
            if not request:
                serv_job.accepting.set()
                conn.sendall(cls._4xx_page(400, "Bad Request", serv_job.auto_close))
                LOG.debug(
                    "400 bad request %r... length %d (%d to go)",
                    raw_request[:16],
                    len(raw_request),
                    serv_job.pending,
                )
                return
            # handle unsupported method
            if request.method != "GET":
                serv_job.accepting.set()
                conn.sendall(
                    cls._4xx_page(405, "Method Not Allowed", serv_job.auto_close)
                )
                LOG.debug("405 method %r (%d to go)", request.method, serv_job.pending)
                return

            # lookup resource
            LOG.debug("lookup resource %r", request.url.path)
            resource = serv_job.lookup_resource(request.url.path)
            if resource:
                if isinstance(resource, FileResource):
                    finish_job = serv_job.remove_pending(str(resource.target))
                elif isinstance(resource, (DynamicResource, RedirectResource)):
                    finish_job = serv_job.remove_pending(request.url.path.lstrip("/"))
                else:  # pragma: no cover
                    # this should never happen
                    raise WorkerError(f"Unknown resource type {resource.type!r}")
                if serv_job.forever:
                    finish_job = False

            if not finish_job:
                serv_job.accepting.set()
            else:
                LOG.debug("expecting to finish")

            # send response
            if resource is None:
                conn.sendall(cls._4xx_page(404, "Not Found", serv_job.auto_close))
                LOG.debug(
                    "404 '%s%s' (%d to go)",
                    "..." if len(request.url.path) > 40 else "",
                    request.url.path[-40:],
                    serv_job.pending,
                )
            elif isinstance(resource, RedirectResource):
                redirect_to = [quote(resource.target)]
                if request.url.query:
                    LOG.debug("appending query %r", request.url.query)
                    redirect_to.append(request.url.query)
                conn.sendall(cls._307_redirect("?".join(redirect_to)))
                LOG.debug(
                    "307 %r -> %r (%d to go)",
                    request.url.path,
                    resource.target,
                    serv_job.pending,
                )
            elif isinstance(resource, DynamicResource):
                # pass query string to callback
                data = resource.target(request.url.query)
                if not isinstance(data, bytes):
                    LOG.debug("dynamic request: %r", request.url.path)
                    raise TypeError("dynamic request callback must return 'bytes'")
                conn.sendall(cls._200_header(len(data), resource.mime))
                conn.sendall(data)
                LOG.debug(
                    "200 %r - dynamic request (%d to go)",
                    request.url.path,
                    serv_job.pending,
                )
            else:
                assert isinstance(resource, FileResource)
                # serve the file
                data_size = resource.target.stat().st_size
                LOG.debug(
                    "sending: %s, %r, '%s'",
                    f"{data_size:,}B",
                    resource.mime,
                    resource.target,
                )
                assert resource.mime is not None
                with resource.target.open("rb") as in_fp:
                    conn.sendall(cls._200_header(data_size, resource.mime))
                    offset = 0
                    while offset < data_size:
                        conn.sendall(in_fp.read(cls.DEFAULT_TX_SIZE))
                        offset = in_fp.tell()
                LOG.debug("200 %r (%d to go)", request.url.path, serv_job.pending)
                serv_job.mark_served(resource)

        except (OSError, sock_timeout):
            _, exc_obj, exc_tb = exc_info()
            LOG.debug("%r - line %r", exc_obj, exc_tb.tb_lineno if exc_tb else None)
            if not finish_job:
                serv_job.accepting.set()

        except Exception:  # pylint: disable=broad-except
            LOG.debug("worker thread exception")
            # set finish_job to abort immediately
            finish_job = True
            if serv_job.exceptions.empty():
                serv_job.exceptions.put(exc_info())

        finally:
            conn.close()
            if finish_job:
                serv_job.finish()
            serv_job.worker_complete.set()

    def join(self, timeout: float = 30) -> bool:
        assert timeout >= 0
        if self._thread is not None:
            self._thread.join(timeout=timeout)
            if not self._thread.is_alive():
                self._thread = None
        return self._thread is None

    @classmethod
    def launch(
        cls, listen_sock: socket, job: Job, timeout: float = 30
    ) -> Worker | None:
        assert timeout >= 0
        assert job.accepting.is_set()
        # TODO: is select() timeout value too short, too long?
        readable, _, _ = select([listen_sock], (), (), 0.25)
        if listen_sock in readable:
            conn = None
            try:
                conn, _ = listen_sock.accept()
                conn.settimeout(timeout)
                # create a worker thread to handle client request
                w_thread = Thread(target=cls.handle_request, args=(conn, job))
                job.accepting.clear()
                w_thread.start()
                return cls(conn, w_thread)
            except BlockingIOError:
                # accept() can block because of race between select() and accept()
                pass
            except OSError as exc:
                LOG.debug("worker thread not launched: %s", exc)
            except ThreadError:
                # reset accepting status
                job.accepting.set()
                LOG.warning("ThreadError (worker), threads: %d", active_count())
                # wait for system resources to free up
                sleep(0.1)
            if conn is not None:
                conn.close()
        return None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from pathlib import Path
from shutil import rmtree

from pytest import raises

from .assets import AssetError, AssetManager


def test_asset_manager_01(tmp_path):
    """test AssetManager()"""
    with AssetManager(base_path=tmp_path) as asset_mgr:
        assert not asset_mgr.assets
        assert asset_mgr.path
        assert asset_mgr.is_empty()
        # add file (move)
        example = tmp_path / "example.txt"
        example.write_text("foo")
        asset_location = asset_mgr.add("example_file", example, copy=False)
        assert asset_mgr.assets["example_file"] == "example.txt"
        assert asset_mgr.path / "example.txt" == asset_location
        assert asset_mgr.get("example_file") == asset_location
        assert len(asset_mgr.assets) == 1
        assert not asset_mgr.is_empty()
        assert not example.is_file()
        assert (asset_mgr.path / "example.txt").is_file()
        # add existing asset - update asset (copy)
        example = tmp_path / "example_2.txt"
        example.write_text("foo")
        asset_mgr.add("example_file", example)
        assert example.is_file()
        assert len(asset_mgr.assets) == 1
        # add directory
        example = tmp_path / "example_path"
        example.mkdir()
        (example / "a").mkdir()
        (example / "a" / "1.txt").write_text("1")
        (example / "b").mkdir()
        (example / "b" / "2.txt").write_text("2")
        asset_mgr.add("example_path", example)
        rmtree(example)
        assert (asset_mgr.path / "example_path/a/1.txt").is_file()
        assert (asset_mgr.path / "example_path/b/2.txt").is_file()
        assert len(asset_mgr.assets) == 2
        # get
        example = asset_mgr.get("example_file")
        assert example
        assert (asset_mgr.path / "example_2.txt").samefile(example)
        example = asset_mgr.get("example_path")
        assert example
        assert (asset_mgr.path / "example_path").samefile(example)
        # remove directory
        asset_mgr.remove("example_path")
        assert len(asset_mgr.assets) == 1
        assert asset_mgr.path.is_dir()
        assert not (asset_mgr.path / "example_path").is_dir()
        # remove file
        asset_mgr.remove("example_file")
        assert len(asset_mgr.assets) == 0
        assert not any(asset_mgr.path.iterdir())
        # cleanup
        asset_mgr.cleanup()
        assert not asset_mgr.assets
        assert not asset_mgr.path.is_dir()


def test_asset_manager_02(tmp_path):
    """test AssetManager() failures"""
    with AssetManager(base_path=tmp_path) as asset_mgr:
        # get missing
        assert asset_mgr.get("missing") is None
        # add missing file
        with raises(OSError, match="'missing' does not exist"):
            asset_mgr.add("a", Path("missing"))
        assert not asset_mgr.assets
        # remove invalid asset
        asset_mgr.remove("missing")
        # add file
        example = tmp_path / "example.txt"
        example.touch()
        asset_mgr.add("example_file", example, copy=True)
        # add with asset with name and file collision
        asset_mgr.add("example_file", example, copy=True)
        # add with file name collision as different asset
        with raises(AssetError, match="collide: 'example.txt' already exists"):
            asset_mgr.add("collide", example)
        assert "collide" not in asset_mgr.assets


def test_asset_manager_03(tmp_path):
    """test AssetManager() load"""
    src = tmp_path / "src"
    src.mkdir()
    (src / "b.txt").touch()

    with AssetManager.load({"a": "b.txt"}, src, base_path=tmp_path) as asset_mgr:
        assert len(asset_mgr.assets) == 1
        assert "a" in asset_mgr.assets
        assert asset_mgr.assets["a"] == "b.txt"
        assert (asset_mgr.path / asset_mgr.assets["a"]).is_file()


def test_asset_manager_04(tmp_path):
    """test AssetManager.add_batch()"""
    batch = []
    with AssetManager(base_path=tmp_path) as asset_mgr:
        # add file
        example = tmp_path / "example.txt"
        example.write_text("example")
        batch.append(["example_file", example])
        # add directory
        example = tmp_path / "example_path"
        example.mkdir()
        (example / "a").mkdir()
        (example / "a" / "1.txt").write_text("1")
        batch.append(["example_path", example])
        asset_mgr.add_batch(batch)
        assert len(asset_mgr.assets) == 2



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Sapphire HTTP server job
"""
from __future__ import annotations

from enum import IntEnum, unique
from errno import ENAMETOOLONG
from itertools import chain
from logging import getLogger
from mimetypes import guess_type
from os.path import splitext
from queue import Queue
from threading import Event, Lock
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Iterable, Mapping, NamedTuple, Tuple, Union, cast

from .server_map import DynamicResource, FileResource, RedirectResource, ServerMap

if TYPE_CHECKING:
    from pathlib import Path

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


@unique
class Served(IntEnum):
    """Server Job status codes"""

    # all expected requests for required files have been received
    ALL = 0
    # no requests for required files have been received
    NONE = 1
    # some requests for required files have been received
    REQUEST = 2
    # timeout occurred
    TIMEOUT = 3


class PendingTracker(NamedTuple):
    files: set[str]
    lock: Lock


class ServedTracker(NamedTuple):
    files: set[FileResource]
    lock: Lock


class Job:
    # MIME_MAP is used to support new or uncommon mime types.
    # Definitions in here take priority over mimetypes.guess_type().
    MIME_MAP = MappingProxyType(
        {
            ".avif": "image/avif",
            ".bmp": "image/bmp",
            ".ico": "image/x-icon",
            ".wave": "audio/x-wav",
            ".webp": "image/webp",
        }
    )

    __slots__ = (
        "_complete",
        "_pending",
        "_served",
        "_wwwroot",
        "auto_close",
        "accepting",
        "exceptions",
        "forever",
        "server_map",
        "worker_complete",
    )

    def __init__(
        self,
        wwwroot: Path,
        auto_close: int = -1,
        forever: bool = False,
        required_files: Iterable[str] | None = None,
        server_map: ServerMap | None = None,
    ) -> None:
        self._complete = Event()
        self._pending = PendingTracker(files=set(), lock=Lock())
        self._served = ServedTracker(files=set(), lock=Lock())
        self._wwwroot = wwwroot.resolve()
        self.accepting = Event()
        self.accepting.set()
        self.auto_close = auto_close
        # quotes around type for Python 3.8
        self.exceptions: Queue[tuple[Any, Any, Any]] = Queue()
        self.forever = forever
        self.server_map = server_map
        self.worker_complete = Event()
        self._build_pending(required_files)
        if not self._pending.files and not self.forever:
            raise RuntimeError("Empty Job")

    def _build_pending(self, required_files: Iterable[str] | None) -> None:
        """Build file list to track files that must be served.
        Note: This is intended to only be called once by __init__().

        Args:
            required_files: File paths (relative to wwwroot) that must be served.

        Returns:
            None
        """
        assert not self._complete.is_set()
        assert not self._pending.files
        assert not self._served.files
        if required_files:
            for required in required_files:
                assert "?" not in required
                entry = self._wwwroot / required
                if entry.is_file():
                    self._pending.files.add(str(entry.resolve()))
                    LOG.debug("required: %r", required)
        # if nothing was found check if the path exists
        if not self._pending.files and not self._wwwroot.is_dir():
            raise OSError(f"wwwroot '{self._wwwroot}' does not exist")
        if self.server_map:
            for url, resource in cast(
                Iterable[Tuple[str, Union[DynamicResource, RedirectResource]]],
                chain(
                    self.server_map.redirect.items(), self.server_map.dynamic.items()
                ),
            ):
                if resource.required:
                    self._pending.files.add(url)
                    LOG.debug("required: %r -> %r", url, resource.target)
        LOG.debug("job has %d required file(s)", len(self._pending.files))

    @classmethod
    def lookup_mime(cls, url: str) -> str:
        """Determine mime type for a given URL.

        Args:
            url: URL to inspect.

        Returns:
            Mime type of URL. 'application/octet-stream' is returned if the mime
            type cannot be determined.
        """
        mime = cls.MIME_MAP.get(splitext(url)[-1].lower())
        if mime is None:
            # default to "application/octet-stream"
            mime = guess_type(url)[0] or "application/octet-stream"
        return mime

    def lookup_resource(
        self, path: str
    ) -> FileResource | DynamicResource | RedirectResource | None:
        """Find the Resource mapped to a given URL path.

        Args:
            path: URL path.

        Returns:
            Resource: Resource for the given URL path or None if one is not found.
        """
        path = path.lstrip("/")
        # check if path is a file in wwwroot
        try:
            local = self._wwwroot / path
            if local.is_file():
                local = local.resolve()
                if self._wwwroot in local.parents:
                    with self._pending.lock:
                        required = str(local) in self._pending.files
                    return FileResource(path, required, local, self.lookup_mime(path))
        except OSError as exc:
            if exc.errno == ENAMETOOLONG:
                # file name is too long to look up so ignore it
                return None
            raise  # pragma: no cover
        # look for path in server map
        if self.server_map is not None:
            if path in self.server_map.redirect:
                return self.server_map.redirect[path]
            if path in self.server_map.dynamic:
                return self.server_map.dynamic[path]
            # search include paths for a match
            for inc in (x for x in self.server_map.include if path.startswith(x)):
                LOG.debug("checking include %r", inc)
                # strip include prefix from potential file name
                file = path[len(inc) :].lstrip("/")
                local = self.server_map.include[inc].target / file
                # check that the file exists within the include path
                if local.is_file():
                    local = local.resolve()
                    if self.server_map.include[inc].target in local.parents:
                        return FileResource(
                            f"{inc}/{file}" if inc else file,
                            self.server_map.include[inc].required,
                            local,
                            self.lookup_mime(file),
                        )
        return None

    def finish(self) -> None:
        """Mark Job as complete.

        Args:
            None

        Returns:
            None
        """
        self._complete.set()

    def mark_served(self, item: FileResource) -> None:
        """Mark a Resource as served to track served Resources.

        Args:
            item: Resource to track.

        Returns:
            None
        """
        with self._served.lock:
            self._served.files.add(item)

    def is_complete(self, wait: float | None = None) -> bool:
        """Check if a Job has been marked as complete.

        Args:
            wait: Time to wait in seconds.

        Returns:
            True if Job complete flag is set otherwise False.
        """
        if wait is not None:
            return self._complete.wait(wait)
        return self._complete.is_set()

    @property
    def pending(self) -> int:
        """Number of files that have not been served.

        Args:
            None

        Returns:
            Number of pending files.
        """
        with self._pending.lock:
            return len(self._pending.files)

    def remove_pending(self, file_name: str) -> bool:
        """Remove a file from pending list.

        Args:
            file_name (str): File to remove from pending list.

        Returns:
            True when all files have been removed otherwise False.
        """
        with self._pending.lock:
            self._pending.files.discard(file_name)
            return not self._pending.files

    @property
    def served(self) -> Mapping[str, Path]:
        """Served files.

        Args:
            None

        Returns:
            Mapping of URLs to files on disk.
        """
        with self._served.lock:
            return MappingProxyType({x.url: x.target for x in self._served.files})

    @property
    def status(self) -> Served:
        """Job Status.

        Args:
            None

        Returns:
            Current status.
        """
        with self._pending.lock:
            if not self._served.files:
                return Served.NONE
            if not self._pending.files:
                return Served.ALL
        return Served.REQUEST



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from dataclasses import dataclass
from inspect import signature
from logging import getLogger
from re import search as re_search
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from pathlib import Path

__all__ = ("Resource", "ServerMap")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


class InvalidURLError(Exception):
    """Raised what a URL string string contained invalid characters"""


class MapCollisionError(Exception):
    """Raised when a URL is already in use by ServerMap"""


@dataclass(frozen=True)
class Resource:
    url: str
    required: bool


@dataclass(frozen=True)
class DynamicResource(Resource):
    target: Callable[[str], bytes]
    mime: str


@dataclass(frozen=True)
class FileResource(Resource):
    target: Path
    mime: str


@dataclass(frozen=True)
class IncludeResource(Resource):
    target: Path


@dataclass(frozen=True)
class RedirectResource(Resource):
    target: str


class ServerMap:
    __slots__ = ("dynamic", "include", "redirect")

    def __init__(self) -> None:
        self.dynamic: dict[str, DynamicResource] = {}
        # mapping of directories that can be requested
        self.include: dict[str, IncludeResource] = {}
        # document paths to map to file names using 307s
        self.redirect: dict[str, RedirectResource] = {}

    @staticmethod
    def _check_url(url: str) -> str:
        # check and sanitize URL
        url = url.strip("/")
        if re_search(r"\W", url) is not None:
            raise InvalidURLError("Only alphanumeric characters accepted in URL.")
        return url

    def set_dynamic_response(
        self,
        url: str,
        callback: Callable[[str], bytes],
        mime_type: str = "application/octet-stream",
        required: bool = False,
    ) -> None:
        url = self._check_url(url)
        if not callable(callback):
            raise TypeError("callback must be callable")
        if len(signature(callback).parameters) != 1:
            raise TypeError("callback requires 1 argument")
        if not isinstance(mime_type, str):
            raise TypeError("mime_type must be of type 'str'")
        if url in self.include or url in self.redirect:
            raise MapCollisionError(f"URL collision on {url!r}")
        LOG.debug("mapping dynamic response %r -> %r (%r)", url, callback, mime_type)
        self.dynamic[url] = DynamicResource(url, required, callback, mime_type)

    def set_include(self, url: str, target: Path) -> None:
        url = self._check_url(url)
        if not target.is_dir():
            raise OSError(f"Include path not found: {target}")
        if url in self.dynamic or url in self.redirect:
            raise MapCollisionError(f"URL collision on {url!r}")
        # sanity check to prevent mapping overlapping paths
        # Note: This was added to help map file served via includes back to
        # the files on disk. This is a temporary workaround until mapping of
        # requests to files that were served is available outside of Sapphire.
        for existing_url, resource in self.include.items():
            if url == existing_url:
                # allow overwriting entry
                continue
            if resource.target in target.parents:
                LOG.error("%r mapping includes path '%s'", existing_url, target)
                raise MapCollisionError(
                    f"{url!r} and {existing_url!r} include '{target}'"
                )
            if target in resource.target.parents:
                LOG.error("%r mapping includes path '%s'", url, resource.target)
                raise MapCollisionError(
                    f"{url!r} and {existing_url!r} include '{resource.target}'"
                )
        LOG.debug("mapping include %r -> '%s'", url, target)
        self.include[url] = IncludeResource(url, False, target)

    def set_redirect(self, url: str, target: str, required: bool = True) -> None:
        url = self._check_url(url)
        if not isinstance(target, str):
            raise TypeError("target must be of type 'str'")
        if not target:
            raise TypeError("target must not be an empty string")
        if url in self.dynamic or url in self.include:
            raise MapCollisionError(f"URL collision on {url!r}")
        self.redirect[url] = RedirectResource(url, required, target)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from enum import IntEnum, unique
from logging import getLogger
from os import environ
from threading import Lock
from typing import TYPE_CHECKING, final

from sapphire import CertificateBundle

from ..common.utils import grz_tmp
from .assets import AssetManager

if TYPE_CHECKING:
    from pathlib import Path

    from ..common.report import Report
    from .target_monitor import TargetMonitor

__all__ = ("Result", "Target", "TargetError", "TargetLaunchError")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]

LOG = getLogger(__name__)


@unique
class Result(IntEnum):
    """Target result codes"""

    NONE = 0
    FOUND = 1
    IGNORED = 2


class TargetError(Exception):
    """Raised by Target"""


class TargetLaunchError(TargetError):
    """Raised if a failure during launch occurs"""

    def __init__(self, message: str, report: Report) -> None:
        super().__init__(message)
        self.report = report


class TargetLaunchTimeout(TargetError):
    """Raised if the target does not launch within the defined amount of time"""


class Target(metaclass=ABCMeta):
    SUPPORTED_ASSETS: tuple[str, ...] = ()
    TRACKED_ENVVARS: tuple[str, ...] = ()

    __slots__ = (
        "_asset_mgr",
        "_https",
        "_lock",
        "_monitor",
        "binary",
        "certs",
        "environ",
        "launch_timeout",
        "log_limit",
        "memory_limit",
    )

    def __init__(
        self,
        binary: Path,
        launch_timeout: int,
        log_limit: int,
        memory_limit: int,
        certs: CertificateBundle | None = None,
    ) -> None:
        assert launch_timeout > 0
        assert log_limit >= 0
        assert memory_limit >= 0
        assert binary is not None and binary.is_file()
        assert certs is None or isinstance(certs, CertificateBundle)
        self._asset_mgr = AssetManager(base_path=grz_tmp("target"))
        self._https = False
        self._lock = Lock()
        self._monitor: TargetMonitor | None = None
        self.binary = binary
        self.certs = certs
        self.environ = self.scan_environment(dict(environ), self.TRACKED_ENVVARS)
        self.launch_timeout = launch_timeout
        self.log_limit = log_limit
        self.memory_limit = memory_limit

    def __enter__(self) -> Target:
        return self

    def __exit__(self, *exc: object) -> None:
        self.cleanup()

    @property
    def asset_mgr(self) -> AssetManager:
        """Get current AssetManager.

        Args:
            None

        Returns:
            AssetManager.
        """
        return self._asset_mgr

    @asset_mgr.setter
    def asset_mgr(self, asset_mgr: AssetManager) -> None:
        """Set AssetManager and cleanup previous AssetManager.

        Args:
            None

        Returns:
            AssetManager.
        """
        self._asset_mgr.cleanup()
        assert isinstance(asset_mgr, AssetManager)
        self._asset_mgr = asset_mgr

    @abstractmethod
    def _cleanup(self) -> None:
        """Cleanup method to be implemented by subclass.

        Args:
            None

        Returns:
            None.
        """

    @abstractmethod
    def check_result(self, ignored: set[str]) -> Result:
        """Check for results.

        Args:
            ignored: Result types that are currently ignored.

        Returns:
            Result code.
        """

    @final
    def cleanup(self) -> None:
        """Perform necessary cleanup. DO NOT OVERRIDE.

        Args:
            ignored: Types of results to ignore.

        Returns:
            Result code.
        """
        # call target specific _cleanup method first
        self._cleanup()
        self._asset_mgr.cleanup()

    @abstractmethod
    def close(self, force_close: bool = False) -> None:
        """Close target.

        Args:
            force_close: Close as quickly as possible. Logs will not be collected.

        Returns:
            None.
        """

    @property
    @abstractmethod
    def closed(self) -> bool:
        """Check if the target is closed.

        Args:
            None

        Returns:
            True if closed otherwise False.
        """

    @abstractmethod
    def create_report(self, is_hang: bool = False, unstable: bool = False) -> Report:
        """Process logs and create a Report.

        Args:
            is_hang: Indicate whether the results is due to a hang/timeout.
            unstable: Indicate whether build is unstable.

        Returns:
            Report object.
        """

    @abstractmethod
    def dump_coverage(self, timeout: int = 0) -> None:
        """Trigger target coverage data dump.

        Args:
            timeout: Amount of time to wait for data to be written.

        Returns:
            None.
        """

    def filtered_environ(self) -> dict[str, str]:
        """Used to collect the environment to add to a testcase.

        Args:
            None

        Returns:
            Environment variables.
        """
        return dict(self.environ)

    @abstractmethod
    def handle_hang(
        self,
        ignore_idle: bool = True,
        ignore_timeout: bool = False,
    ) -> bool:
        """Handle a target hang.

        Args:
            ignore_idle: Do not treat as a hang if target is idle.
            ignore_timeout: Indicates if a timeout will be ignored.

        Returns:
            True if the target was idle otherwise False.
        """

    @abstractmethod
    def https(self) -> bool:
        """Target configured for HTTPS.

        Args:
            None

        Returns:
            True if HTTPS can be used otherwise False.
        """

    @abstractmethod
    def launch(self, location: str) -> None:
        """Launch the target.

        Args:
            location: URL to load.

        Returns:
            None.
        """

    @abstractmethod
    def log_size(self) -> int:
        """Calculate the amount of data contained in target log files.

        Args:
            None

        Returns:
            Total data size of log files in bytes.
        """

    @abstractmethod
    def merge_environment(self, extra: dict[str, str]) -> None:
        """Add to existing environment.

        Args:
            extra: Environment variables to add.

        Returns:
            None.
        """

    @property
    @abstractmethod
    def monitor(self) -> TargetMonitor:
        """TargetMonitor.

        Args:
            extra: Environment variables to add.

        Returns:
            TargetMonitor
        """

    @abstractmethod
    def process_assets(self) -> None:
        """Prepare assets for use by the target.

        Args:
            None

        Returns:
            None.
        """

    def reverse(self, remote: int, local: int) -> None:
        """Configure port mappings. Remote -> device, local -> desktop (current system).

        Args:
            remote: Port on remote device.
            local: Port on local machine.

        Returns:
            None.
        """

    @staticmethod
    def scan_environment(
        env: dict[str, str],
        include: tuple[str, ...] | None,
    ) -> dict[str, str]:
        """Scan environment for tracked environment variables.

        Args:
            env: Environment to scan.
            include: Variables to include in output.

        Returns:
            Tracked variables found in scanned environment.
        """
        return {var: env[var] for var in include if var in env} if include else {}

    @abstractmethod
    def save_logs(self, dst: Path) -> None:
        """Save logs to specified location.

        Args:
            dst: Location to save logs.

        Returns:
            None.
        """



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from time import perf_counter, sleep
from typing import Callable

from sapphire import Sapphire, Served, ServerMap

from ..target import Result, Target, TargetLaunchError, TargetLaunchTimeout
from .storage import TestCase

__all__ = ("Runner", "RunResult")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)
# time in seconds to wait for target to navigate away from post launch page
POST_LAUNCH_TIMEOUT = 600
# display warning if launch duration exceeds set value
SLOW_LAUNCH_THRESHOLD = 20


class _IdleChecker:
    """_IdleChecker is used to help determine if the target is hung (actively using CPU)
    or if it has not made expected the HTTP requests for other reasons (idle).
    This will allow the framework to move on without interrupting execution of long
    running test cases.
    This is not perfect! It is to be used AFTER the test case timeout (initial_delay)
    has elapsed.
    """

    __slots__ = ("_check_cb", "_init_delay", "_poll_delay", "_threshold", "_next_poll")

    def __init__(
        self,
        check_cb: Callable[[int], bool],
        threshold: int,
        initial_delay: int,
        poll_delay: float = 1,
    ) -> None:
        assert callable(check_cb)
        assert initial_delay >= 0
        assert poll_delay >= 0
        assert 100 > threshold >= 0
        self._check_cb = check_cb  # callback used to check if target is idle
        self._init_delay = initial_delay  # time to wait before the initial idle poll
        self._poll_delay = poll_delay  # time to wait between subsequent polls
        self._threshold = threshold  # CPU usage threshold
        self._next_poll: float | None = None

    def is_idle(self) -> bool:
        """Check the target idle callback. This is throttled by '_next_poll'
        which specifies the time at which the next callback call is allowed.

        Args:
            None

        Returns:
            True if the target idle callback returned True otherwise False
        """
        assert self._next_poll is not None, "schedule_poll() must be called first"
        now = perf_counter()
        if now >= self._next_poll:
            if self._check_cb(self._threshold):
                return True
            self.schedule_poll(now=now)
        return False

    def schedule_poll(self, initial: int = False, now: float | None = None) -> None:
        """Update `_next_poll`.

        Args:
            initial: Use `_init_delay` to schedule next poll.
            now: Time in seconds.

        Returns:
            None
        """
        if now is None:
            now = perf_counter()
        if initial:
            self._next_poll = now + self._init_delay
        else:
            self._next_poll = now + self._poll_delay


@dataclass(eq=False)
class RunResult:
    """A RunResult holds result details from a call to Runner.run().

    Attributes:
        attempted: Test entry point was requested.
        duration: Time spent waiting for test contents to be served.
        served: Files that were served.
        status: Result status of test.
        timeout: A timeout occurred waiting for test to complete.
        idle: Target was idle (only applies to timeout).
    """

    served: tuple[str, ...]
    duration: float
    attempted: bool = False
    status: Result = Result.NONE
    timeout: bool = False
    idle: bool = False


class Runner:
    __slots__ = (
        "_close_delay",
        "_idle",
        "_relaunch",
        "_server",
        "_target",
        "_tests_run",
        "startup_failure",
    )

    def __init__(
        self,
        server: Sapphire,
        target: Target,
        close_delay: int = 30,
        idle_threshold: int = 0,
        idle_delay: int = 0,
        relaunch: int = 1,
    ) -> None:
        self._close_delay = close_delay
        if idle_threshold > 0:
            assert idle_delay > 0
            LOG.debug("using idle check, th %d, delay %ds", idle_threshold, idle_delay)
            self._idle: _IdleChecker | None = _IdleChecker(
                target.monitor.is_idle, idle_threshold, idle_delay
            )
        else:
            self._idle = None
        assert close_delay > 0
        assert relaunch > 0
        self._relaunch = relaunch  # tests to run before relaunching target
        self._server = server  # a sapphire instance to serve the test case
        self._target = target  # target to run test case
        self._tests_run = 0  # number of tests run since target (re)launched
        self.startup_failure = False  # failure before first test was served

    def launch(self, location: str, max_retries: int = 3, retry_delay: int = 0) -> None:
        """Launch a target and open `location`.

        Args:
            location: URL to open via Target.
            max_retries: Number of retries to perform before re-raising
                         TargetLaunchTimeout.
            retry_delay: Time in seconds to wait between retries.

        Returns:
            None
        """
        assert self._server is not None
        assert self._target is not None
        assert max_retries >= 0
        assert retry_delay >= 0
        # nothing should be trying to connect, did the previous target.close() fail?
        assert self._server.clear_backlog()
        self._tests_run = 0
        self.startup_failure = False
        launch_duration: float = 0
        LOG.debug("launching target (timeout %ds)", self._target.launch_timeout)
        for retries in reversed(range(max_retries)):
            try:
                launch_start = perf_counter()
                self._target.launch(location)
                launch_duration = perf_counter() - launch_start
            except (TargetLaunchError, TargetLaunchTimeout) as exc:
                is_timeout = isinstance(exc, TargetLaunchTimeout)
                if not retries:
                    self.startup_failure = True
                    msg = "timeout" if is_timeout else "failed"
                    LOG.error(
                        "Launch %s, please verify browser build works as expected", msg
                    )
                    raise
                if is_timeout:
                    # A TargetLaunchTimeout likely has nothing to do with Grizzly but is
                    # seen frequently on machines under a high load. After multiple
                    # consecutive timeouts something is likely wrong so raise.
                    LOG.warning("Timeout during launch (retries %d)", retries)
                else:
                    assert isinstance(exc, TargetLaunchError)
                    # This is likely due to a bad build or environment configuration.
                    LOG.warning("Failure during launch (retries %d)", retries)
                    exc.report.cleanup()
                sleep(retry_delay)
                continue
            break

        if launch_duration > SLOW_LAUNCH_THRESHOLD:
            LOG.warning(
                "Slow launch detected (%0.1fs > %ds)",
                launch_duration,
                SLOW_LAUNCH_THRESHOLD,
            )

    @staticmethod
    def location(
        srv_path: str,
        srv_port: int,
        close_after: int | None = None,
        post_launch_delay: int = -1,
        scheme: str = "http",
        time_limit: int | None = None,
    ) -> str:
        """Build a valid URL to pass to a browser.

        Args:
            srv_path: Path segment of the URL
            srv_port: Server listening port
            close_after: Harness argument.
            post_launch_delay: Post-launch delay page argument.
            scheme: URL scheme component (http or https).
            time_limit: Harness argument.

        Returns:
            A valid URL.
        """
        location = f"{scheme}://localhost:{srv_port}/{srv_path.lstrip('/')}"
        # set harness related arguments
        args = []
        if close_after is not None:
            assert close_after >= 0
            args.append(f"close_after={close_after}")
        if time_limit:
            assert time_limit > 0
            args.append(f"time_limit={time_limit * 1000}")
        if post_launch_delay >= 0:
            args.append(f"post_launch_delay={post_launch_delay}")
        if args:
            return "?".join((location, "&".join(args)))
        return location

    @property
    def initial(self) -> bool:
        """Check if more than one test has been run since the previous target launch.

        Args:
            None

        Returns:
            True if at most one test has been run.
        """
        return self._tests_run < 2

    def post_launch(self, delay: int = 0) -> None:
        """Perform actions after launching browser before loading test cases.

        Args:
            delay: Time in seconds before the target will continue.

        Returns:
            None
        """
        assert delay >= 0
        with TestCase("post_launch_delay.html", "None") as content:
            content.add_from_file(
                Path(__file__).parent / "post_launch_delay.html",
                file_name=content.entry_point,
                copy=True,
            )
            srv_map = ServerMap()
            srv_map.set_redirect("grz_start", content.entry_point, required=False)
            srv_map.set_redirect("grz_continue", "grz_start", required=True)
            # temporarily override server timeout
            org_timeout = self._server.timeout
            # add time buffer to redirect delay
            # in practice this should take a few seconds (~10s)
            # in extreme cases ~180s (slow build + debugger + other settings)
            self._server.timeout = delay + POST_LAUNCH_TIMEOUT
            if delay > 0:
                LOG.info("Browser launched, continuing in %ds...", delay)
            LOG.debug("post launch timeout: %ds", self._server.timeout)
            # serve prompt page
            server_status, _ = self._server.serve_path(
                content.root,
                continue_cb=self._target.monitor.is_healthy,
                server_map=srv_map,
            )
            # restore server timeout
            self._server.timeout = org_timeout
            if server_status != Served.ALL:
                self.startup_failure = True
                if server_status == Served.TIMEOUT:
                    # this should never happen with a correctly functioning build
                    LOG.warning("Target hung after launch")
                LOG.warning("Post launch check failed!")

    def run(
        self,
        ignore: set[str],
        server_map: ServerMap,
        testcase: TestCase,
        coverage: bool = False,
        wait_for_callback: bool = False,
    ) -> RunResult:
        """Serve a testcase and monitor the target for results.

        Args:
            ignore: Failure types to ignore.
            server_map: A ServerMap.
            testcase: The test case that will be served.
            coverage: Trigger coverage dump.
            wait_for_callback: Use `_keep_waiting()` to indicate when framework
                               should move on.

        Returns:
            Files served, status and timeout flag from the run.
        """
        self._tests_run += 1
        if self._idle is not None:
            self._idle.schedule_poll(initial=True)
        if self._tests_run == self._relaunch:
            # overwrite instead of replace 'grz_next_test' for consistency
            server_map.set_redirect("grz_next_test", "grz_empty", required=True)
            server_map.set_dynamic_response("grz_empty", lambda _: b"", required=True)
        # clear optional contents from test case
        # it will be repopulated with served contests
        testcase.clear_optional()
        # serve the test case
        serve_start = perf_counter()
        server_status, served = self._server.serve_path(
            testcase.root,
            continue_cb=self._keep_waiting,
            forever=wait_for_callback,
            required_files=tuple(testcase.required),
            server_map=server_map,
        )
        duration = perf_counter() - serve_start
        result = RunResult(
            tuple(served),
            duration,
            attempted=testcase.entry_point in served,
            timeout=server_status == Served.TIMEOUT,
        )
        # add all files that were served (includes, etc...) to test
        existing = set(testcase.required)
        for url, local_file in served.items():
            if url not in existing:
                # copy include files
                testcase.add_from_file(local_file, file_name=url, copy=True)
        # record use of https in testcase
        testcase.https = self._server.scheme == "https"
        if result.timeout:
            LOG.debug("timeout detected")
            result.idle = self._target.handle_hang(
                ignore_idle=True, ignore_timeout="timeout" in ignore
            )
            if result.idle or "timeout" in ignore:
                result.status = Result.IGNORED
            server_map.dynamic.pop("grz_empty", None)
        if result.attempted:
            if coverage and not result.timeout:
                # dump_coverage() should be called before check_result()
                # to help catch any coverage related issues.
                self._target.dump_coverage()
            # relaunch check
            if self._tests_run >= self._relaunch and not result.timeout:
                assert self._tests_run == self._relaunch
                server_map.dynamic.pop("grz_empty", None)
                LOG.debug("relaunch/shutdown limit hit")
                # ideally all browser tabs should be closed at this point
                # and the browser should exit on its own
                # NOTE: this will take the full duration if target.monitor.is_idle()
                # is unable to detect if the target is idle
                for close_delay in range(max(int(self._close_delay / 0.5), 1)):
                    if not self._target.monitor.is_healthy():
                        break
                    # wait 3 seconds (6 passes) before attempting idle exit
                    if close_delay > 5 and self._target.monitor.is_idle(10):
                        # NOTE: this will always trigger on systems where the
                        # browser does not exit when the last window is closed
                        LOG.debug("target idle")
                        break
                    # delay to help catch shutdown related crashes, LSan, etc.
                    # debugger and different builds can slow shutdown
                    sleep(0.5)
                else:
                    LOG.debug("target.close() required")
                self._target.close()
        else:
            # something is wrong so close the target
            # previous iteration put target in a bad state?
            LOG.debug("entry point not served (%r)", testcase.entry_point)
            self._target.close()
            # detect startup failures
            if self.initial:
                self.startup_failure = True
        # detect results
        if result.status == Result.NONE:
            result.status = self._target.check_result(ignore)
        return result

    def _keep_waiting(self) -> bool:
        """Callback used by the server to determine if it should continue to
        wait for the requests from the target.

        Args:
            None

        Returns:
            Continue to serve the test case.
        """
        if self._idle is not None and self._idle.is_idle():
            LOG.debug("idle target detected")
            return False
        return self._target.monitor.is_healthy()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
unit tests for grizzly.replay.main
"""
from pathlib import Path
from unittest.mock import Mock

from pytest import mark

from sapphire import Served

from ..common.report import Report
from ..common.storage import TestCase, TestCaseLoadFailure
from ..common.utils import Exit
from ..target import (
    AssetManager,
    Result,
    Target,
    TargetLaunchError,
    TargetLaunchTimeout,
)
from .replay import ReplayManager
from .test_replay import _fake_save_logs

pytestmark = mark.usefixtures(
    "patch_collector",
    "tmp_path_grz_tmp",
)


def test_main_01(mocker, server, tmp_path):
    """test ReplayManager.main()"""
    # This is a typical scenario - a test that reproduces results ~50% of the time.
    # Of the four attempts only the first and third will 'reproduce' the result
    # and the forth attempt should be skipped.
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    server.serve_path.return_value = (Served.ALL, {"test.html": "/fake/path"})
    # setup Target
    load_target = mocker.patch("grizzly.replay.replay.load_plugin", autospec=True)
    target = mocker.Mock(
        spec_set=Target, binary=Path("bin"), environ={}, launch_timeout=30
    )
    target.asset_mgr = mocker.Mock(spec_set=AssetManager)
    target.check_result.side_effect = (Result.FOUND, Result.NONE, Result.FOUND)
    target.filtered_environ.return_value = {"ENV": "123"}
    target.save_logs = _fake_save_logs
    load_target.return_value.return_value = target
    with TestCase("test.html", "adpt") as src:
        src.env_vars["TEST_VAR"] = "100"
        src.add_from_bytes(b"test", src.entry_point)
        src.dump(tmp_path / "testcase", include_details=True)
    # setup args
    log_path = tmp_path / "logs"
    (tmp_path / "sig.json").write_bytes(
        b'{"symptoms": [{"type": "crashAddress", "address": "0"}]}'
    )
    args = mocker.Mock(
        any_crash=False,
        asset=[],
        entry_point=None,
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=["fake", "timeout"],
        input=[tmp_path / "testcase"],
        launch_attempts=3,
        min_crashes=2,
        no_harness=False,
        output=log_path,
        pernosco=False,
        post_launch_delay=0,
        relaunch=1,
        repeat=4,
        rr=False,
        sig=tmp_path / "sig.json",
        test_index=[],
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    assert ReplayManager.main(args) == Exit.SUCCESS
    assert target.reverse.call_count == 1
    assert target.launch.call_count == 3
    assert target.check_result.call_count == 3
    assert target.merge_environment.call_count == 1
    assert server.serve_path.call_count == 6
    assert load_target.call_count == 1
    assert target.close.call_count == 4
    assert target.filtered_environ.call_count == 2
    assert target.cleanup.call_count == 1
    assert target.asset_mgr.add.call_count == 0
    assert target.asset_mgr.is_empty.call_count == 1
    assert log_path.is_dir()
    assert any(log_path.glob("reports/*/log_asan_blah.txt"))
    assert any(log_path.glob("reports/*/log_stderr.txt"))
    assert any(log_path.glob("reports/*/log_stdout.txt"))


@mark.parametrize(
    "repro_results,",
    [
        # no results
        (Result.NONE, Result.NONE),
        # results do not match signature
        (Result.FOUND, Result.FOUND),
        # no result and failed signature match
        (Result.FOUND, Result.NONE),
    ],
)
def test_main_02(mocker, server, tmp_path, repro_results):
    """test ReplayManager.main() - no repro"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    server.serve_path.return_value = (Served.ALL, {"test.html": "/fake/path"})
    # setup Target
    target = mocker.Mock(
        spec_set=Target, binary=Path("bin"), environ={}, launch_timeout=30
    )
    target.check_result.side_effect = repro_results
    target.filtered_environ.return_value = {}
    target.save_logs = _fake_save_logs
    mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        return_value=mocker.Mock(spec=Target, return_value=target),
    )
    (tmp_path / "sig.json").write_bytes(
        b'{"symptoms": [{"type": "stackFrames", "functionNames": ["no-match"]}]}'
    )
    # setup args
    (tmp_path / "test.html").touch()
    args = mocker.Mock(
        any_crash=False,
        asset=[],
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=[],
        input=[tmp_path / "test.html"],
        launch_attempts=3,
        min_crashes=2,
        no_harness=True,
        output=tmp_path / "logs",
        pernosco=False,
        post_launch_delay=-1,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=tmp_path / "sig.json",
        test_index=[],
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    assert ReplayManager.main(args) == Exit.FAILURE
    assert target.check_result.call_count == 1
    assert not target.environ
    assert target.close.call_count == 2
    assert target.cleanup.call_count == 1


@mark.parametrize(
    "load_plugin, load_testcases, signature, result",
    [
        # failed to load test cases
        (None, TestCaseLoadFailure("test"), None, Exit.ERROR),
        # user abort
        (
            KeyboardInterrupt,
            ([Mock(spec_set=TestCase, hang=False)], None, None),
            None,
            Exit.ABORT,
        ),
        # signature required replaying hang
        (None, ([Mock(spec_set=TestCase)], None, None), None, Exit.ERROR),
        # can't ignore timeout replaying hang
        (None, ([Mock(spec_set=TestCase)], None, None), "sig", Exit.ERROR),
        # cleanup assets loaded from test case
        (
            KeyboardInterrupt,
            ([Mock(spec_set=TestCase, hang=False)], Mock(spec_set=AssetManager), None),
            None,
            Exit.ABORT,
        ),
    ],
)
def test_main_03(mocker, load_plugin, load_testcases, signature, result):
    """test ReplayManager.main() error cases"""
    mocker.patch("grizzly.replay.replay.Sapphire", autospec=True)
    mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        side_effect=load_plugin,
    )
    mocker.patch(
        "grizzly.replay.replay.ReplayManager.load_testcases",
        autospec=True,
        side_effect=(load_testcases,),
    )
    fake_sig = mocker.patch("grizzly.replay.replay.CrashSignature", autospec=True)
    # setup args
    args = mocker.Mock(
        asset=[],
        fuzzmanager=False,
        ignore=["timeout"] if signature else [],
        input=["test"],
        min_crashes=1,
        pernosco=False,
        post_launch_delay=-1,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=signature,
        test_index=[],
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    asset_mgr = load_testcases[1] if isinstance(load_testcases, tuple) else None
    assert ReplayManager.main(args) == result
    if asset_mgr is not None:
        assert asset_mgr.cleanup.call_count == 1
    if signature:
        assert fake_sig.fromFile.call_count == 1


def test_main_04(mocker, tmp_path):
    """test ReplayManager.main() target exceptions"""
    mocker.patch("grizzly.replay.replay.FuzzManagerReporter", autospec=True)
    reporter = mocker.patch("grizzly.replay.replay.FailedLaunchReporter", autospec=True)
    mocker.patch("grizzly.replay.replay.Sapphire", autospec=True)
    mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        return_value=mocker.Mock(
            spec_set=Target,
            return_value=mocker.NonCallableMock(spec_set=Target, launch_timeout=30),
        ),
    )
    with TestCase("test.html", "adpt") as test:
        test.add_from_bytes(b"", test.entry_point)
        test.dump(tmp_path / "test", include_details=True)
    # setup args
    args = mocker.MagicMock(
        entry_point=None,
        input=[tmp_path / "test"],
        min_crashes=1,
        no_harness=True,
        post_launch_delay=-1,
        pernosco=False,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=None,
        test_index=[],
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    # target launch error
    mocker.patch(
        "grizzly.replay.replay.ReplayManager.run",
        side_effect=TargetLaunchError("", mocker.Mock(spec_set=Report)),
    )
    assert ReplayManager.main(args) == Exit.LAUNCH_FAILURE
    assert reporter.return_value.submit.call_count == 1
    reporter.reset_mock()
    # target launch timeout
    mocker.patch(
        "grizzly.replay.replay.ReplayManager.run", side_effect=TargetLaunchTimeout
    )
    assert ReplayManager.main(args) == Exit.LAUNCH_FAILURE
    assert reporter.return_value.submit.call_count == 0


def test_main_05(mocker, server, tmp_path):
    """test ReplayManager.main() loading specified assets"""
    server.serve_path.return_value = (None, {"test.html": "/fake/path"})
    # setup Target
    target = mocker.NonCallableMock(
        spec_set=Target, binary=Path("bin"), launch_timeout=30
    )
    target.check_result.return_value = Result.FOUND
    target.filtered_environ.return_value = {}
    target.monitor.is_healthy.return_value = False
    target.save_logs = _fake_save_logs
    mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        return_value=mocker.Mock(spec_set=Target, return_value=target),
    )
    asset = tmp_path / "sample_asset"
    asset.touch()
    input_path = tmp_path / "input"
    input_path.mkdir()
    log_path = tmp_path / "logs"
    # setup args
    args = mocker.Mock(
        asset=[["from_cmdline", str(asset)]],
        entry_point=None,
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=[],
        input=[input_path],
        launch_attempts=3,
        min_crashes=1,
        no_harness=True,
        output=log_path,
        pernosco=False,
        post_launch_delay=-1,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=None,
        test_index=[],
        time_limit=1,
        timeout=None,
        valgrind=False,
    )
    # build a test case
    with TestCase("test.html", "test-adapter") as src:
        src.add_from_bytes(b"test", src.entry_point)
        src.dump(input_path, include_details=True)
    with AssetManager(base_path=tmp_path) as asset_mgr:
        target.asset_mgr = asset_mgr
        assert ReplayManager.main(args) == Exit.SUCCESS
        assert "from_cmdline" in target.asset_mgr.assets
    assert target.launch.call_count == 1
    assert target.check_result.call_count == 1
    assert target.filtered_environ.call_count == 1
    assert server.serve_path.call_count == 1
    assert log_path.is_dir()
    assert any(log_path.glob("**/sample_asset"))


@mark.parametrize(
    "pernosco, rr, valgrind, no_harness",
    [
        # No debugger enabled and no harness
        (False, False, False, False),
        # No debugger enabled and with harness
        (False, False, False, True),
        # Pernosco enabled
        (True, False, False, False),
        # rr enabled
        (False, True, False, False),
        # Valgrind enabled
        (False, False, True, False),
    ],  # pylint: disable=invalid-name
)
def test_main_06(
    mocker, server, tmp_path, pernosco, rr, valgrind, no_harness
):  # pylint: disable=invalid-name
    """test ReplayManager.main() enable debuggers"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    server.serve_path.return_value = (Served.ALL, {"test.html": "/fake/path"})
    # setup Target
    target = mocker.NonCallableMock(spec_set=Target, binary="bin", launch_timeout=30)
    target.check_result.return_value = Result.NONE
    load_target = mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        return_value=mocker.Mock(spec_set=Target, return_value=target),
    )
    # setup args
    (tmp_path / "test.html").touch()
    args = mocker.Mock(
        asset=[],
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=["fake", "timeout"],
        input=[tmp_path / "test.html"],
        launch_attempts=3,
        min_crashes=2,
        no_harness=no_harness,
        pernosco=pernosco,
        post_launch_delay=-1,
        relaunch=1,
        repeat=1,
        rr=rr,
        sig=None,
        test_index=[],
        time_limit=10,
        timeout=None,
        valgrind=valgrind,
    )
    # maximum one debugger allowed at a time
    assert sum((pernosco, rr, valgrind)) < 2, "test broken!"
    assert ReplayManager.main(args) == Exit.FAILURE
    assert target.check_result.call_count == 1
    assert target.close.call_count == 2
    assert target.cleanup.call_count == 1
    assert load_target.return_value.call_args[-1]["pernosco"] == pernosco
    assert load_target.return_value.call_args[-1]["rr"] == rr
    assert load_target.return_value.call_args[-1]["valgrind"] == valgrind


def test_main_07(mocker, server, tmp_path):
    """test ReplayManager.main() - report to FuzzManager"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    server.serve_path.return_value = (Served.ALL, {"test.html": "/fake/path"})
    reporter = mocker.patch("grizzly.replay.replay.FuzzManagerReporter", autospec=True)
    # setup Target
    load_target = mocker.patch("grizzly.replay.replay.load_plugin", autospec=True)
    target = mocker.MagicMock(
        spec_set=Target, binary=Path("bin"), environ={}, launch_timeout=30
    )
    target.check_result.side_effect = (Result.FOUND,)
    target.save_logs = _fake_save_logs
    load_target.return_value.return_value = target
    with TestCase("test.html", "adpt") as src:
        src.add_from_bytes(b"test", src.entry_point)
        src.dump(tmp_path / "testcase", include_details=True)
    # setup args
    args = mocker.Mock(
        any_crash=False,
        asset=[],
        entry_point=None,
        fuzzmanager=True,
        idle_delay=0,
        idle_threshold=0,
        ignore=[],
        input=[tmp_path / "testcase"],
        launch_attempts=1,
        min_crashes=1,
        no_harness=False,
        output=None,
        pernosco=False,
        post_launch_delay=-1,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=None,
        test_index=[],
        time_limit=10,
        timeout=None,
        tool=None,
        valgrind=False,
    )
    assert ReplayManager.main(args) == Exit.SUCCESS
    assert target.reverse.call_count == 1
    assert target.launch.call_count == 1
    assert target.check_result.call_count == 1
    assert server.serve_path.call_count == 1
    assert load_target.call_count == 1
    assert target.close.call_count == 2
    assert target.cleanup.call_count == 1
    assert reporter.return_value.submit.call_count == 1


@mark.parametrize("https_supported", [True, False])
def test_main_08(mocker, tmp_path, https_supported):
    """test ReplayManager.main() - Target HTTPS support"""
    (tmp_path / "test.html").touch()
    args = mocker.MagicMock(
        adapter="fake",
        binary=Path("bin"),
        input=[tmp_path / "test.html"],
        fuzzmanager=False,
        launch_attempts=1,
        min_crashes=1,
        relaunch=1,
        repeat=1,
        sig=None,
        use_http=False,
        time_limit=1,
        timeout=1,
    )

    target_cls = mocker.MagicMock(spec_set=Target)
    target = target_cls.return_value
    target.https.return_value = https_supported
    mocker.patch("grizzly.replay.replay.load_plugin", return_value=target_cls)
    mocker.patch("grizzly.replay.replay.ReplayManager.run", return_value=[])
    assert ReplayManager.main(args) == Exit.FAILURE
    assert target.https.call_count == 1


def test_main_09(mocker, server, tmp_path):
    """test ReplayManager.main() - load test case assets"""
    server.serve_path.return_value = (None, {"test.html": "/fake/path"})
    # setup Target
    target = mocker.NonCallableMock(
        spec_set=Target, binary=Path("bin"), launch_timeout=30
    )
    target.check_result.return_value = Result.NONE
    target.monitor.is_healthy.return_value = False
    mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        return_value=mocker.Mock(spec_set=Target, return_value=target),
    )
    # setup args
    input_path = tmp_path / "input"
    input_path.mkdir()
    args = mocker.MagicMock(
        adapter="fake",
        binary=Path("bin"),
        entry_point=None,
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        input=[input_path],
        launch_attempts=1,
        min_crashes=1,
        output=None,
        post_launch_delay=-1,
        relaunch=1,
        repeat=1,
        sig=None,
        use_http=False,
        test_index=[],
        time_limit=1,
        timeout=1,
    )
    # build test case and asset
    asset = tmp_path / "sample_asset"
    asset.touch()
    with AssetManager(base_path=tmp_path) as asset_mgr:
        asset_mgr.add("sample", asset)
        with TestCase("test.html", "test-adapter") as src:
            src.assets = asset_mgr.assets
            src.assets_path = asset_mgr.path
            src.add_from_bytes(b"", src.entry_point)
            src.dump(input_path, include_details=True)
    # this will load the previously created test case and asset from the filesystem
    try:
        assert ReplayManager.main(args) == Exit.FAILURE
        assert target.launch.call_count == 1
        assert "sample" in target.asset_mgr.assets
        assert target.asset_mgr.path is not None
    finally:
        if target.asset_mgr:
            target.asset_mgr.cleanup()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly reducer lithium strategy definitions."""
from __future__ import annotations

from abc import ABC
from logging import getLogger
from typing import TYPE_CHECKING, Iterator

from lithium.strategies import CheckOnly, Minimize, ReductionIterator
from lithium.strategies import CollapseEmptyBraces as LithCollapseEmptyBraces
from lithium.strategies import Strategy as LithStrategy
from lithium.testcases import Testcase as LithTestcase
from lithium.testcases import TestcaseAttrs, TestcaseChar, TestcaseJsStr, TestcaseLine

from ...common.storage import TestCase
from . import Strategy, _contains_dd

if TYPE_CHECKING:
    from pathlib import Path

LOG = getLogger(__name__)


class _LithiumStrategy(Strategy, ABC):
    """Use a Lithium `Strategy`/`Testcase` pair to reduce the given Grizzly `TestCase`
    set.

    Implementers must define these class attributes:

    Class attributes:
        name: The strategy name.
        strategy_cls: Lithium strategy type.
        testcase_cls: Lithium testcase type.
    """

    strategy_cls: type[LithStrategy]
    testcase_cls: type[LithTestcase]

    def __init__(self, testcases: list[TestCase]) -> None:
        """Initialize strategy instance.

        Arguments:
            testcases: Testcases to reduce. The object does not take ownership of the
                       testcases.
        """
        super().__init__(testcases)
        self._current_feedback: bool | None = None
        self._current_reducer: ReductionIterator | None = None
        self._files_to_reduce: list[Path] = []
        self.rescan_files_to_reduce(
            [
                TestCase.load(x, catalog=True)
                for x in sorted(self._testcase_root.iterdir())
            ]
        )

    def rescan_files_to_reduce(self, testcases: list[TestCase]) -> None:
        """Repopulate the private `files_to_reduce` attribute by scanning the testcase
        root.

        Returns:
            None
        """
        self._files_to_reduce.clear()
        for test in testcases:
            for path in (test.root / x for x in test):
                if _contains_dd(path):
                    self._files_to_reduce.append(path)

    def update(self, success: bool) -> None:
        """Inform the strategy whether or not the last reduction yielded was good.

        Arguments:
            success: Whether or not the last reduction was acceptable.

        Returns:
            None
        """
        if self._current_reducer is not None:
            self._current_reducer.feedback(success)
        self._current_feedback = success

    def __iter__(self) -> Iterator[list[TestCase]]:
        """Iterate over potential reductions of testcases according to this strategy.

        The caller should evaluate each testcase set yielded, and call `update` with the
        result. The caller owns the testcases yielded, and should call `cleanup` for
        each.

        Yields:
            Testcases with reduction applied.
        """
        LOG.info("Reducing %d files", len(self._files_to_reduce))
        file_no = 0
        reduce_queue = self._files_to_reduce.copy()
        reduce_queue.sort()  # not necessary, but helps make tests more predictable
        while reduce_queue:
            LOG.debug(
                "Reduce queue: %r",
                ", ".join(
                    str(x.relative_to(self._testcase_root)) for x in reduce_queue
                ),
            )
            file = reduce_queue.pop(0)
            file_no += 1
            LOG.info(
                "[%s] Reducing %s (file %d/%d)",
                self.name,
                file.relative_to(self._testcase_root),
                file_no,
                len(self._files_to_reduce),
            )
            lithium_testcase = self.testcase_cls()  # pylint: disable=not-callable
            lithium_testcase.load(file)
            strategy = self.strategy_cls()  # pylint: disable=not-callable
            self._current_reducer = strategy.reduce(lithium_testcase)

            # Populate the lithium strategy "tried" cache.
            # Use all cache values where all hashes other than the current file match
            #   the current testcase_root state.
            current_tc_hash_map = dict(self._calculate_testcase_hash())
            del current_tc_hash_map[str(file.relative_to(self._testcase_root))]
            this_tc_tried = set()
            for tried in self._tried:
                tried_map = dict(tried)
                tc_tried = tried_map.pop(str(file.relative_to(self._testcase_root)))
                if tried_map == current_tc_hash_map:
                    this_tc_tried.add(tc_tried)
            self._current_reducer.update_tried(this_tc_tried)

            for reduction in self._current_reducer:
                reduction.dump()
                testcases = [
                    TestCase.load(x, catalog=True)
                    for x in sorted(self._testcase_root.iterdir())
                ]
                LOG.info("[%s] %s", self.name, self._current_reducer.description)
                yield testcases
                if not self._current_feedback:
                    self._tried.add(self._calculate_testcase_hash())
                else:
                    LOG.debug(
                        "files being reduced before: %r",
                        ", ".join(
                            str(x.relative_to(self._testcase_root))
                            for x in self._files_to_reduce
                        ),
                    )
                    self.rescan_files_to_reduce(testcases)
                    LOG.debug(
                        "files being reduced after: %r",
                        ", ".join(
                            str(x.relative_to(self._testcase_root))
                            for x in self._files_to_reduce
                        ),
                    )
                    files_to_reduce = set(self._files_to_reduce)
                    reduce_queue = sorted(set(reduce_queue) & files_to_reduce)
                    if file not in files_to_reduce:
                        # current reduction was for a purged file
                        break
            else:
                # write out the best found testcase
                self._current_reducer.testcase.dump()
            self._current_reducer = None


class Check(_LithiumStrategy):
    """Check whether the testcase is reproducible.

    This strategy does no reduction, and only yields once. It is intended to provide a
    pass/fail result in a reduction pipeline.
    """

    name = "check"
    strategy_cls = CheckOnly
    testcase_cls = TestcaseLine

    def __init__(self, testcases: list[TestCase]) -> None:
        super().__init__(testcases)
        # trim files_to_reduce, for check we don't need to run on every file
        # just once per Grizzly TestCase set is enough.
        self._files_to_reduce = self._files_to_reduce[:1]

    def __iter__(self) -> Iterator[list[TestCase]]:
        yield from super().__iter__()


class CollapseEmptyBraces(_LithiumStrategy):
    """Minimize lines, but collapse empty curly braces between each iteration.

    During reduction, the contents of a block may be reduced away entirely, but removing
    the starting brace or ending brace alone will break the syntax of the test file.
    This strategy tries to collapse empty braces onto the same line between each
    iteration, so that empty blocks can be removed if otherwise possible.
    """

    name = "collapsebraces"
    strategy_cls = LithCollapseEmptyBraces
    testcase_cls = TestcaseLine


class MinimizeChars(_LithiumStrategy):
    """Minimize all bytes in the testcase."""

    name = "chars"
    strategy_cls = Minimize
    testcase_cls = TestcaseChar


class MinimizeJSChars(_LithiumStrategy):
    """Minimize all bytes contained in javascript strings.

    This works the same as MinimizeChars, but only operates if it can identify what
    looks like a quoted string. It also treats escaped characters as a single token
    for reduction.
    """

    name = "jschars"
    strategy_cls = Minimize
    testcase_cls = TestcaseJsStr


class MinimizeLines(_LithiumStrategy):
    """Minimize all lines in the testcase."""

    name = "lines"
    strategy_cls = Minimize
    testcase_cls = TestcaseLine


class MinimizeAttrs(_LithiumStrategy):
    """Remove HTML/XML attributes and their values."""

    name = "attrs"
    strategy_cls = Minimize
    testcase_cls = TestcaseAttrs



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Callable

from ..common.fuzzmanager import Bucket
from ..common.utils import Exit, configure_logging
from .args import ReplayFuzzManagerIDQualityArgs
from .crash import main as crash_main

if TYPE_CHECKING:
    from argparse import Namespace

LOG = getLogger(__name__)


def bucket_main(args: Namespace, tool_main: Callable[[Namespace], int]) -> int:
    """CLI for `grizzly.reduce.bucket` and `grizzly.replay.bucket`.

    Arguments:
        args: Result from `parse_args()`.
        tool_main: Main function from a supported Grizzly tool.

    Returns:
        Exit.SUCCESS (0) for success otherwise a different Exit code is returned.
    """
    assert callable(tool_main)
    configure_logging(args.log_level)
    LOG.info("Trying each crash in bucket %d until successful", args.input)

    # ensure --tool is reset for each call to tool_main()
    orig_tool = args.tool

    # if no crashes in bucket, return success
    result = Exit.SUCCESS

    with Bucket(args.input) as bucket:
        if args.sig is None:
            args.sig = bucket.signature_path()

        for crash in bucket.iter_crashes(args.quality):
            args.input = crash.crash_id
            args.tool = orig_tool

            # call grizzly tool
            result = Exit(tool_main(args))
            if result in {Exit.SUCCESS, Exit.ABORT}:
                break

    return result


def main() -> int:
    """Wrapper for bucket_main() which is the CLI for `grizzly.replay.bucket`.

    Arguments:
        None

    Returns:
        Exit.SUCCESS (0) for success otherwise a different Exit code is returned.
    """
    return bucket_main(ReplayFuzzManagerIDQualityArgs().parse_args(), crash_main)


if __name__ == "__main__":
    raise SystemExit(main())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from tempfile import mkdtemp
from typing import TYPE_CHECKING, Callable, cast

from FTB.Signatures.CrashInfo import CrashSignature

from sapphire import CertificateBundle, Sapphire, ServerMap

from ..common.plugins import load_plugin
from ..common.report import Report
from ..common.reporter import (
    FailedLaunchReporter,
    FilesystemReporter,
    FuzzManagerReporter,
)
from ..common.runner import Runner, RunResult
from ..common.status import SimpleStatus
from ..common.storage import TestCase, TestCaseLoadFailure
from ..common.utils import (
    HARNESS_FILE,
    ConfigError,
    Exit,
    configure_logging,
    display_time_limits,
    grz_tmp,
    package_version,
    time_limits,
)
from ..target import (
    AssetManager,
    Result,
    Target,
    TargetLaunchError,
    TargetLaunchTimeout,
)
from .args import ReplayArgs

if TYPE_CHECKING:
    from argparse import Namespace

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


@dataclass(eq=False)
class ReplayResult:
    """Contains information related to the replay of testcases.

    Attributes:
        report: Report containing logs.
        durations: Number of seconds spent running each testcase.
        expected: Signature match.
        count: Number of times detected.
    """

    report: Report
    durations: list[float]
    expected: bool
    count: int = 1


class ReplayManager:
    __slots__ = (
        "ignore",
        "server",
        "status",
        "target",
        "_any_crash",
        "_harness",
        "_signature",
        "_relaunch",
        "_unpacked",
    )

    def __init__(
        self,
        ignore: set[str],
        server: Sapphire,
        target: Target,
        any_crash: bool = False,
        relaunch: int = 1,
        signature: CrashSignature | None = None,
        use_harness: bool = True,
    ) -> None:
        # target must relaunch every iteration when not using harness
        assert use_harness or relaunch == 1
        self.ignore = ignore
        self.server = server
        self.status: SimpleStatus | None = None
        self.target = target
        self._any_crash = any_crash
        self._harness = HARNESS_FILE.read_bytes() if use_harness else None
        self._relaunch = relaunch
        self._signature = signature

    def __enter__(self) -> ReplayManager:
        return self

    def __exit__(self, *exc: object) -> None:
        self.cleanup()

    @property
    def signature(self) -> CrashSignature | None:
        return self._signature

    @staticmethod
    def check_match(
        signature: CrashSignature, report: Report, expect_hang: bool, check_failed: bool
    ) -> bool:
        """Check if report matches signature.

        Args:
            signature: Signature to match.
            report: Report to check against signature.
            expect_hang: Indicates if a hang is expected.
            check_failed: Check if report signature creation failed.

        Returns:
            True if report matches signature otherwise False.
        """
        if signature is None:
            if check_failed and not report.is_hang:
                # treat failed signature creation as a match
                return report.crash_signature is None
            # treat 'None' signature as a bucket if it's not a hang
            return not report.is_hang
        if expect_hang and not report.is_hang:
            # avoid catching other crashes with forgiving hang signatures
            return False
        # Fuzzmanager is missing type hints, use cast()
        return cast(bool, signature.matches(report.crash_info))

    def cleanup(self) -> None:
        """Remove temporary files from disk.

        Args:
            None

        Returns:
            None
        """
        # TODO: Do we need this anymore?

    @staticmethod
    def expect_hang(
        ignore: list[str], signature: CrashSignature, tests: list[TestCase]
    ) -> bool:
        """Check if any test is expected to trigger a hang. If a hang is expected
        a sanity check is performed. A ConfigError is raised if a configuration
        issue is detected.

        Args:
            ignore: Failure types to ignore.
            signature: Signature to use for bucketing.
            tests: Testcases to check.

        Returns:
            True if a hang is expected otherwise False.
        """
        is_hang = any(x.hang for x in tests)
        if is_hang:
            if signature is None:
                raise ConfigError("Hangs require a signature to replay", Exit.ERROR)
            if "timeout" in ignore:
                raise ConfigError(
                    "Cannot ignore 'timeout' when detecting hangs", Exit.ERROR
                )
        return is_hang

    @classmethod
    def load_testcases(
        cls,
        paths: list[Path],
        catalog: bool = False,
        entry_point: Path | None = None,
    ) -> tuple[list[TestCase], AssetManager | None, dict[str, str] | None]:
        """Load TestCases.

        Args:
            paths: Testcases to load.
            catalog: See TestCase.load().
            entry_point: See TestCase.load().

        Returns:
            Loaded TestCases, AssetManager and environment variables.
        """
        LOG.debug("loading the TestCases")
        tests: list[TestCase] = []
        for entry in paths:
            try:
                tests.append(
                    TestCase.load(entry, catalog=catalog, entry_point=entry_point)
                )
            except TestCaseLoadFailure as exc:  # noqa: PERF203
                LOG.warning("Failed to load: '%s' (%s)", entry, exc)
        if not tests:
            raise TestCaseLoadFailure("Failed to load TestCases")
        # load and remove assets and environment variables from test cases
        asset_mgr = None
        env_vars = None
        for test in tests:
            if asset_mgr is None and test.assets and test.assets_path:
                asset_mgr = AssetManager.load(test.assets, test.assets_path)
            if not env_vars and test.env_vars:
                env_vars = dict(test.env_vars)
            test.env_vars.clear()
            test.assets.clear()
            test.assets_path = None
        LOG.debug(
            "loaded TestCase(s): %d, assets: %r, env vars: %r",
            len(tests),
            asset_mgr is not None,
            env_vars is not None,
        )
        return tests, asset_mgr, env_vars

    @staticmethod
    def lookup_tool(tests: list[TestCase]) -> str | None:
        """Lookup tool name from test cases. Find the adapter name used by the given
        test cases.

        Args:
            tests: TestCase to scan.

        Returns:
            TestCase adapter name or None.
        """
        adapter_name = {x.adapter_name for x in tests if x.adapter_name}
        if adapter_name:
            return f"grizzly-{adapter_name.pop()}"
        return None

    @staticmethod
    def report_to_filesystem(
        dst: Path, results: list[ReplayResult], tests: list[TestCase]
    ) -> None:
        """Use FilesystemReporter to write reports and testcase to disk in a
        known location.

        Args:
            dst: Location to write data.
            results: ReplayResult to output.
            tests: Testcases to output.

        Returns:
            None
        """
        others = tuple(x.report for x in results if not x.expected)
        if others:
            reporter = FilesystemReporter(dst / "other_reports", major_bucket=False)
            for report in others:
                reporter.submit(tests or [], report)
        expected = tuple(x.report for x in results if x.expected)
        if expected:
            reporter = FilesystemReporter(dst / "reports", major_bucket=False)
            for report in expected:
                reporter.submit(tests or [], report)

    @staticmethod
    def report_to_fuzzmanager(
        results: list[ReplayResult], tests: list[TestCase], tool: str
    ) -> None:
        """Use FuzzManagerReporter to send reports to a FuzzManager server.

        Args:
            results: ReplayResult to output.
            tests: Testcases to output.
            tool: Name used by FuzzManager.

        Returns:
            None
        """
        reporter = FuzzManagerReporter(tool)
        for result in results:
            # always report expected results
            # avoid reporting unexpected frequent results
            reporter.submit(tests, result.report, force=result.expected)

    def run(
        self,
        testcases: list[TestCase],
        time_limit: int,
        repeat: int = 1,
        min_results: int = 1,
        exit_early: bool = True,
        expect_hang: bool = False,
        idle_delay: int = 0,
        idle_threshold: int = 0,
        launch_attempts: int = 3,
        on_iteration_cb: Callable[[], None] | None = None,
        post_launch_delay: int = -1,
    ) -> list[ReplayResult]:
        """Run testcase replay.

        Args:
            testcases: One or more TestCases to run.
            time_limit: Maximum time in seconds a test should take to complete.
            repeat: Maximum number of times to run the TestCase.
            min_results: Minimum number of results needed before run can be considered
                         successful.
            expect_hang: Running testcases is expected to result in a hang.
            exit_early: If True the minimum required number of iterations are performed
                        to either meet `min_results` or determine that it is not
                        possible to do so. If False `repeat` number of iterations are
                        performed.
            idle_delay: Number of seconds to wait before polling for idle.
            idle_threshold: CPU usage threshold to mark the process as idle.
            launch_attempts: Number of attempts to launch the browser.
            on_iteration_cb: Called every time a single iteration is run.
            post_launch_delay: Number of seconds to wait before continuing after the
                               browser is launched. A negative number skips redirect.

        Returns:
            ReplayResults that were found running provided testcases.
        """
        assert idle_delay >= 0
        assert idle_threshold >= 0
        assert launch_attempts > 0
        assert min_results > 0
        assert repeat > 0
        assert repeat >= min_results
        assert testcases
        assert time_limit > 0 or self._harness is None
        assert self._harness is not None or len(testcases) == 1
        assert not expect_hang or self._signature is not None

        self.status = SimpleStatus.start()

        server_map = ServerMap()
        if self._harness is None:
            server_map.set_redirect("grz_start", "grz_current_test", required=False)
        else:
            assert isinstance(self._harness, bytes)
            assert self._harness, "harness must contain data"

            def harness_fn(_: str) -> bytes:  # pragma: no cover
                assert self._harness
                return self._harness

            server_map.set_dynamic_response(
                "grz_harness",
                harness_fn,
                mime_type="text/html",
            )
            server_map.set_redirect("grz_start", "grz_harness", required=False)

        # track unprocessed results
        reports: dict[str, ReplayResult] = {}
        try:
            sig_hash = Report.calc_hash(self._signature) if self._signature else None
            # an attempt has been made to set self._signature
            sig_set = self._signature is not None
            test_count = len(testcases)
            relaunch = min(self._relaunch, repeat)
            runner = Runner(
                self.server,
                self.target,
                idle_threshold=idle_threshold,
                idle_delay=idle_delay,
                relaunch=relaunch * test_count,
            )
            # perform iterations
            for _ in range(repeat):
                self.status.iteration += 1
                if on_iteration_cb is not None:
                    on_iteration_cb()
                if self.target.closed:
                    location = runner.location(
                        "/grz_start",
                        self.server.port,
                        close_after=relaunch * test_count if self._harness else None,
                        post_launch_delay=post_launch_delay,
                        scheme=self.server.scheme,
                        time_limit=time_limit if self._harness else None,
                    )
                    runner.launch(location, max_retries=launch_attempts)
                    if post_launch_delay >= 0 and not runner.startup_failure:
                        runner.post_launch(delay=post_launch_delay)
                    # TODO: avoid running test case if runner.startup_failure is True
                # run tests
                durations: list[float] = []
                run_result: RunResult | None = None
                for test_idx in range(test_count):
                    if test_count > 1:
                        LOG.info(
                            "Running test, part %d/%d (%d/%d)...",
                            test_idx + 1,
                            test_count,
                            self.status.iteration,
                            repeat,
                        )
                    else:
                        LOG.info(
                            "Running test (%d/%d)...", self.status.iteration, repeat
                        )
                    # update redirects
                    if self._harness is not None:
                        next_idx = (test_idx + 1) % test_count
                        server_map.set_redirect(
                            "grz_next_test",
                            testcases[next_idx].entry_point,
                            required=True,
                        )
                    server_map.set_redirect(
                        "grz_current_test",
                        testcases[test_idx].entry_point,
                        required=False,
                    )
                    # run testcase
                    run_result = runner.run(
                        self.ignore,
                        server_map,
                        testcases[test_idx],
                        wait_for_callback=self._harness is None,
                    )
                    durations.append(run_result.duration)
                    if run_result.status != Result.NONE or not run_result.attempted:
                        break
                assert run_result is not None
                if not run_result.attempted:
                    LOG.warning("Test case was not served")
                    if run_result.timeout:
                        LOG.warning("Browser hung? Timeout too short? System too busy?")
                    elif runner.initial:
                        # TODO: what is the best action to take in this case?
                        LOG.warning("Delayed startup failure detected")
                # process run results
                if run_result.status == Result.FOUND:
                    report: Report | None = None
                    # processing the result may take a few minutes (rr)
                    # update console to show progress
                    LOG.info("Processing result...")
                    # TODO: use self.target.create_report here
                    log_path = Path(mkdtemp(prefix="logs_", dir=grz_tmp("logs")))
                    self.target.save_logs(log_path)
                    report = Report(
                        log_path,
                        self.target.binary,
                        is_hang=run_result.timeout,
                        unstable=runner.startup_failure,
                    )
                    # set active signature
                    if not self._any_crash and not run_result.timeout and not sig_set:
                        assert not expect_hang
                        assert self._signature is None
                        LOG.debug(
                            "no signature given, using short sig %r",
                            report.short_signature,
                        )
                        if runner.startup_failure:
                            LOG.warning(
                                "Using signature from startup failure! "
                                "Provide a signature to avoid this."
                            )
                        self._signature = report.crash_signature
                        sig_set = True
                        if self._signature is not None:
                            assert not sig_hash, "sig_hash should only be set once"
                            sig_hash = Report.calc_hash(self._signature)
                    # look for existing buckets (signature match)
                    expected = self._any_crash or self.check_match(
                        self._signature, report, expect_hang, sig_set
                    )
                    if expected:
                        if sig_hash is not None:
                            LOG.debug("using signature hash (%s) to bucket", sig_hash)
                            bucket_hash = sig_hash
                        else:
                            bucket_hash = report.crash_hash
                        self.status.results.count(bucket_hash, report.short_signature)
                    else:
                        bucket_hash = report.crash_hash
                        self.status.ignored += 1
                    LOG.info(
                        "%s: %s (%s:%s)",
                        "Result" if expected else "Result: Different signature",
                        report.short_signature,
                        report.major[:8],
                        report.minor[:8],
                    )
                    # bucket result
                    if bucket_hash not in reports:
                        reports[bucket_hash] = ReplayResult(report, durations, expected)
                        LOG.debug("now tracking %s", bucket_hash)
                        report = None  # don't remove report
                    else:
                        reports[bucket_hash].count += 1
                        if report.unstable and not reports[bucket_hash].report.unstable:
                            LOG.debug("updating report to unstable")
                            reports[bucket_hash].report.unstable = True
                        LOG.debug("already tracking %s", bucket_hash)
                    # purge untracked report
                    if report is not None:
                        report.cleanup()
                        report = None
                elif run_result.status == Result.IGNORED:
                    self.status.ignored += 1
                    if run_result.timeout:
                        LOG.info(
                            "Result: Ignored (%d); timeout, idle: %r",
                            self.status.ignored,
                            run_result.idle,
                        )
                    else:
                        LOG.info("Result: Ignored (%d)", self.status.ignored)

                if exit_early and self.status.iteration < repeat:
                    # check if failed to meet minimum number of results
                    if (
                        repeat - self.status.iteration + self.status.results.total
                        < min_results
                    ):
                        # failed to reproduce issue
                        LOG.debug("skipping remaining attempts")
                        LOG.debug(
                            "results (%d) < minimum (%d), after %d attempts",
                            self.status.results.total,
                            min_results,
                            self.status.iteration,
                        )
                        # NOTE: this can be tricky if the harness is used because it can
                        # skip the shutdown performed in the harness and runner, if this
                        # is an issue for now use relaunch=1
                        break
                    # check if complete (minimum number of results found)
                    if self.status.results.total >= min_results:
                        assert self.status.results.total == min_results
                        assert (
                            sum(x.count for x in reports.values() if x.expected)
                            >= min_results
                        )
                        LOG.debug(
                            "results == expected (%d), after %d attempts",
                            min_results,
                            self.status.iteration,
                        )
                        break

                # TODO: should we warn about large browser logs?

            # process results
            if self._any_crash:
                # all reports should be expected when self._any_crash=True
                assert all(x.expected for x in reports.values())
                success = sum(x.count for x in reports.values()) >= min_results
                if not success:
                    LOG.debug(
                        "%d (any_crash) less than minimum %d",
                        self.status.results.total,
                        min_results,
                    )
            else:
                # there should be at most one expected bucket
                assert sum(x.expected for x in reports.values()) <= 1
                success = any(
                    x.count >= min_results for x in reports.values() if x.expected
                )
            results: list[ReplayResult] = []
            for crash_hash, result in reports.items():
                # if min_results not met (success=False) cleanup expected reports
                if not success and result.expected:
                    if not self._any_crash:
                        LOG.debug(
                            "%r less than minimum (%d/%d)",
                            crash_hash,
                            result.count,
                            min_results,
                        )
                    result.report.cleanup()
                    continue
                results.append(result)

            # this should only be displayed when both conditions are met:
            # 1) runner does not close target (no delay was given before shutdown)
            # 2) result has not been successfully reproduced
            if (
                self._relaunch > 1
                and not self.target.closed
                and not any(x.expected for x in results)
            ):
                LOG.info("Perhaps try with --relaunch=1")
            # active reports have been moved to results
            # clear reports to avoid cleanup of active reports
            reports.clear()
            return results

        finally:
            # we don't want to cleanup but we are not checking results
            self.target.close(force_close=True)
            # remove unprocessed reports
            for result in reports.values():
                result.report.cleanup()

    @classmethod
    def main(cls, args: Namespace | None = None) -> int:
        """CLI for `grizzly.reduce`.

        Arguments:
            args: Result from `ReplayArgs.parse_args`.

        Returns:
            Exit.SUCCESS (0) for success otherwise a different Exit code is returned.
        """
        args = args or ReplayArgs().parse_args()
        configure_logging(args.log_level)

        LOG.info("Starting Grizzly Replay")
        LOG.debug("grizzly-framework version: %s", package_version("grizzly-framework"))

        if args.headless:
            LOG.info("Running browser headless (%s)", args.headless)
        if args.ignore:
            LOG.info("Ignoring: %s", ", ".join(args.ignore))
        if args.pernosco:
            LOG.info("Running with RR (Pernosco mode)")
        elif args.rr:
            LOG.info("Running with RR")
        elif args.valgrind:
            LOG.info("Running with Valgrind. This will be SLOW!")

        signature = CrashSignature.fromFile(args.sig) if args.sig else None

        try:
            testcases, asset_mgr, env_vars = cls.load_testcases(
                args.input, entry_point=args.entry_point
            )
        except TestCaseLoadFailure as exc:
            LOG.error("Error: %s", str(exc))
            return Exit.ERROR

        if not args.tool:
            args.tool = cls.lookup_tool(testcases) or "grizzly-replay"

        certs = None
        results: list[ReplayResult] | None = None
        target: Target | None = None
        try:
            # check if hangs are expected
            expect_hang = cls.expect_hang(args.ignore, signature, testcases)
            # calculate test time limit and timeout
            time_limit, timeout = time_limits(
                args.time_limit, args.timeout, tests=testcases
            )
            display_time_limits(time_limit, timeout, args.no_harness)
            # calculate repeat and relaunch
            repeat = max(args.min_crashes, args.repeat)
            relaunch = min(args.relaunch, repeat)
            LOG.info(
                "Repeat: %d, Minimum crashes: %d, Relaunch %d",
                repeat,
                args.min_crashes,
                relaunch,
            )
            if not args.use_http:
                certs = CertificateBundle.create()
            LOG.debug("initializing the Target")
            target = load_plugin(args.platform, "grizzly_targets", Target)(
                args.binary,
                args.launch_timeout,
                args.log_limit,
                args.memory,
                certs=certs,
                headless=args.headless,
                pernosco=args.pernosco,
                rr=args.rr,
                valgrind=args.valgrind,
            )
            assert target
            if env_vars is not None:
                LOG.debug("adding environment loaded from test case")
                target.merge_environment(env_vars)

            # use asset manager created from test case content if available
            if asset_mgr:
                target.asset_mgr = asset_mgr
                # target is now responsible for `asset_mgr`
                asset_mgr = None
            # TODO: prioritize specified assets over included
            target.asset_mgr.add_batch(args.asset)
            target.process_assets()

            if certs and not target.https():
                LOG.warning("Target does not support HTTPS, using HTTP")
                certs.cleanup()
                certs = None

            LOG.debug("starting sapphire server")
            # launch HTTP server used to serve test cases
            with Sapphire(auto_close=1, timeout=timeout, certs=certs) as server:
                target.reverse(server.port, server.port)
                with cls(
                    set(args.ignore),
                    server,
                    target,
                    any_crash=args.any_crash,
                    relaunch=relaunch,
                    signature=signature,
                    use_harness=not args.no_harness,
                ) as replay:
                    results = replay.run(
                        testcases,
                        time_limit,
                        expect_hang=expect_hang,
                        idle_delay=args.idle_delay,
                        idle_threshold=args.idle_threshold,
                        launch_attempts=args.launch_attempts,
                        min_results=args.min_crashes,
                        post_launch_delay=args.post_launch_delay,
                        repeat=repeat,
                    )
            # handle results
            success = any(x.expected for x in results)
            if success:
                LOG.info("Results successfully reproduced")
            elif any(results):
                LOG.info("Results detected, signature does not match")
            else:
                LOG.info("No results detected")
            if results and (args.output or args.fuzzmanager):
                # add target assets to test cases
                if not target.asset_mgr.is_empty():
                    for test in testcases:
                        test.assets = dict(target.asset_mgr.assets)
                        test.assets_path = target.asset_mgr.path
                # add target environment variables
                if target.filtered_environ():
                    for test in testcases:
                        test.env_vars = target.filtered_environ()
                # report results
                if args.fuzzmanager:
                    cls.report_to_fuzzmanager(results, testcases, args.tool)
                else:
                    cls.report_to_filesystem(args.output, results, testcases)
            return Exit.SUCCESS if success else Exit.FAILURE

        except ConfigError as exc:
            LOG.error(str(exc))
            return exc.exit_code

        except KeyboardInterrupt:
            return Exit.ABORT

        except (TargetLaunchError, TargetLaunchTimeout) as exc:
            if isinstance(exc, TargetLaunchError) and exc.report:
                FailedLaunchReporter(args.display_launch_failures).submit(
                    [], exc.report
                )
            return Exit.LAUNCH_FAILURE

        finally:
            LOG.info("Shutting down...")
            if results:
                # cleanup unreported results
                for result in results:
                    result.report.cleanup()
            if target is not None:
                target.cleanup()
            if asset_mgr:
                asset_mgr.cleanup()
            if certs is not None:
                certs.cleanup()
            LOG.info("Done.")
